@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2004, 2005, 2008, 2009
+ * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2004, 2005, 2008, 2009,
+ *               2010, 2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -41,11 +42,10 @@
  *
  */
 
-/* $Id: socks.h,v 1.239.4.4 2010/09/14 07:46:36 karls Exp $ */
+/* $Id: socks.h,v 1.256 2011/07/20 14:16:08 karls Exp $ */
 
 #ifndef _SOCKS_H_
 #define _SOCKS_H_
-#endif /* !_SOCKS_H_ */
 
 #ifndef HAVE_OSF_OLDSTYLE
 #define HAVE_OSF_OLDSTYLE 0
@@ -366,14 +366,18 @@
 #define FDPASS_MAX         2   /* max number of descriptors we send/receive.  */
 
 struct configstate_t {
-   int               init;             /* inited?                             */
-   sig_atomic_t      insignal;         /* executing in signalhandler?         */
-
-   struct sockaddr   lastconnect;      /* address we last connected to.       */
-   pid_t             pid;              /* our pid.                            */
-   int               havegssapisockets;/* have gssapi-sockets?                */
-   rlim_t            maxopenfiles;
+   unsigned char      inited;
+   sig_atomic_t       insignal;         /* executing in signalhandler?        */
+   sig_atomic_t       signalforus;      /*
+                                        * between now and the time this
+                                        * variable was last cleared, did we
+                                        * handle a signal?
+                                        */
 
+   struct sockshost_t lastconnect;      /* address we last connected to.      */
+   pid_t              pid;              /* our pid.                           */
+   int                havegssapisockets;/* have gssapi-sockets?               */
+   rlim_t             maxopenfiles;
 };
 
 struct option_t {
@@ -384,15 +388,23 @@ struct option_t {
 
 struct config_t {
    pid_t                    connectchild;            /* connect process.      */
-   int                      child_data;              /* datasocket to child.  */
+   int                      child_data;              /* data socket to child.  */
    int                      child_ack;               /* ack to child.         */
 
    char                     domain[MAXHOSTNAMELEN];  /* localdomain.          */
+
+   struct logtype_t         errlog;                  /* for errors only.      */
    struct logtype_t         log;                     /* where to log.         */
+   int                      loglock;                 /* lockfile for logging. */
+
    struct option_t          option;                  /* misc. options.        */
-   struct configstate_t     state;
    int                      resolveprotocol;         /* resolveprotocol.      */
+
+   routeoptions_t           routeoptions;            /* global route flags.   */
    struct route_t           *route;                  /* linked list of routes */
+
+   struct configstate_t     state;
+   struct timeout_t         timeout;
 };
 
 struct childpacket_t {
@@ -412,14 +424,6 @@ clientinit(void);
  * initializes client state, reads configfile, etc.
  */
 
-void upnpcleanup(const int s);
-/*
- * cleanup upnp-stuff related to the socket "s", mostly involving removal
- * of port mappings.
- * If "s" is -1, clean up for all known sockets.
- */
-
-
 #if !HAVE_OSF_OLDSTYLE
 int Raccept(int, struct sockaddr *, socklen_t *);
 int Rconnect(int, const struct sockaddr *, socklen_t);
@@ -486,7 +490,7 @@ struct route_t *
 udpsetup(int s, const struct sockaddr *to, int type);
 /*
  * sets up udp relaying between address of "s" and "to" by connecting
- * to a proxyserver.
+ * to a proxy server.
  * If relaying is already set up the function returns with success.
  * Type is the type of connection to set up, SOCKS_SEND or SOCKS_RECV.
  *
@@ -497,7 +501,11 @@ udpsetup(int s, const struct sockaddr *to, int type);
     *  Misc. functions to help keep track of our connection(s) to the server.
     */
 
-void addrlockinit(void);
+void socks_addrinit(void);
+/*
+ * inits thing, including memory and locks, for socks_addaddr()-functions.
+ */
+
 void socks_addrlock(const int locktype, addrlockopaque_t *opaque);
 void socks_addrunlock(const addrlockopaque_t *opaque);
 /*
@@ -525,7 +533,7 @@ socks_addaddr(const int clientfd, const struct socksfd_t *socksaddress,
  * socksfdv/addrlock.
  *
  * The function duplicates all arguments in it's own form and does
- * not access the memory referenced by them afterwards.
+ * not access the memory referenced by them afterwords.
  *
  * The function checks the state of all file descriptors on each call and
  * removes those that are no longer open.
@@ -537,8 +545,12 @@ socks_addaddr(const int clientfd, const struct socksfd_t *socksaddress,
  */
 
 struct socksfd_t *
-socks_getaddr(const int fd, const int takelock);
+socks_getaddr(const int fd, struct socksfd_t *socksfd, const int takelock);
 /*
+ * Returns a copy of the socksfd corresponding to "fd".
+ * If "socksfd" is not NULL, the contents of the socksfd is also stored in
+ * "socksfd".
+ *
  * If "takelock" is true, it means the function should take the
  * socksfdv/addrlock.
  *
@@ -605,13 +617,17 @@ socks_isaddr(const int fd, const int takelock);
  */
 
 int
-socks_addrisours(const int s, const int takelock);
+socks_addrisours(const int s, struct socksfd_t *socksfd, const int takelock);
 /*
+ * Compares the current address of "s" to the registered address.
+ * If there is a mismatch, the function will try to correct it if possible.
+ *
  * If "takelock" is true, it means the function should take the
  * socksfdv/addrlock.
  *
- * Compares the current address of "s" to the registered address.
- * If there is a mismatch, the function will try to correct it if possible.
+ * If the current address matches the registered address and "socksfd"
+ * is not NULL, "socksfd" is filled in with the data of the matching socket.
+ *
  * Returns:
  *      If current address found to match registered: true.
  *      Else: false.
@@ -686,10 +702,6 @@ sys_getpeername(HAVE_PROT_GETPEERNAME_1, HAVE_PROT_GETPEERNAME_2,
       HAVE_PROT_GETPEERNAME_3);
 #endif /* HAVE_EXTRA_OSF_SYMBOLS */
 
-HAVE_PROT_GETSOCKOPT_0
-sys_getsockopt(HAVE_PROT_GETSOCKOPT_1, HAVE_PROT_GETSOCKOPT_2,
-      HAVE_PROT_GETSOCKOPT_3, HAVE_PROT_GETSOCKOPT_4, HAVE_PROT_GETSOCKOPT_5);
-
 #if HAVE_OSF_OLDSTYLE
 int sys_getsockname(int, struct sockaddr *, int *);
 #else
@@ -698,6 +710,10 @@ sys_getsockname(HAVE_PROT_GETSOCKNAME_1, HAVE_PROT_GETSOCKNAME_2,
       HAVE_PROT_GETSOCKNAME_3);
 #endif /* HAVE_EXTRA_OSF_SYMBOLS */
 
+HAVE_PROT_GETSOCKOPT_0
+sys_getsockopt(HAVE_PROT_GETSOCKOPT_1, HAVE_PROT_GETSOCKOPT_2,
+      HAVE_PROT_GETSOCKOPT_3, HAVE_PROT_GETSOCKOPT_4, HAVE_PROT_GETSOCKOPT_5);
+
 #if HAVE_OSF_OLDSTYLE
 int sys_recvfrom(int, void*, int, int, struct sockaddr *, int *);
 #else
@@ -844,4 +860,25 @@ sys_fread(HAVE_PROT_FREAD_1, HAVE_PROT_FREAD_2, HAVE_PROT_FREAD_3,
       HAVE_PROT_FREAD_4);
 #endif /* HAVE_GSSAPI && HAVE_LINUX_GLIBC_WORKAROUND */
 
+#if HAVE_DARWIN
+
+HAVE_PROT_READ_0
+sys_read_nocancel(HAVE_PROT_READ_1, HAVE_PROT_READ_2, HAVE_PROT_READ_3);
+HAVE_PROT_CONNECT_0
+sys_connect_nocancel(HAVE_PROT_CONNECT_1, HAVE_PROT_CONNECT_2,
+   HAVE_PROT_CONNECT_3 namelen);
+HAVE_PROT_RECVFROM_0
+sys_recvfrom_nocancel(HAVE_PROT_RECVFROM_1, HAVE_PROT_RECVFROM_2,
+   HAVE_PROT_RECVFROM_3, HAVE_PROT_RECVFROM_4, HAVE_PROT_RECVFROM_5,
+   HAVE_PROT_RECVFROM_6);
+HAVE_PROT_SENDTO_0
+sys_sendto_nocancel(HAVE_PROT_SENDTO_1, HAVE_PROT_SENDTO_2, HAVE_PROT_SENDTO_3,
+   HAVE_PROT_SENDTO_4, HAVE_PROT_SENDTO_5, HAVE_PROT_SENDTO_6);
+HAVE_PROT_WRITE_0
+sys_write_nocancel(HAVE_PROT_WRITE_1, HAVE_PROT_WRITE_2, HAVE_PROT_WRITE_3);
+
+#endif /* HAVE_DARWIN */
+
 #endif /* SOCKSLIBRARY_DYNAMIC */
+
+#endif /* !_SOCKS_H_ */
