@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2003, 2006, 2008, 2009
+ * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2003, 2006, 2008, 2009, 2010,
+ *               2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -41,11 +42,10 @@
  *
  */
 
-/* $Id: tostring.h,v 1.32 2009/10/23 11:08:02 karls Exp $ */
+/* $Id: tostring.h,v 1.44 2011/08/01 09:04:36 michaels Exp $ */
 
 #ifndef _TOSTRING_H_
 #define _TOSTRING_H_
-#endif /* !_TOSTRING_H_ */
 
 #if HAVE_DUMPCONF
 #define QUOTE(a)   __CONCAT3("\"", a, "\"")
@@ -103,7 +103,6 @@ commands2string(const struct command_t *command, char *str, size_t strsize)
 
 char *
 methods2string(size_t methodc, const int *methodv, char *str, size_t strsize)
-      __attribute__((__nonnull__(3)))
       __attribute__((__bounded__(__string__, 3, 4)));
 /*
  * Returns a printable representation of the methods "methodv", of
@@ -207,14 +206,17 @@ string2udpheader(const char *data, size_t len, struct udpheader_t *header)
  * "data" is assumed to be in network order.
  * Returns:
  *      On success: pointer to a udpheader_t in static memory.
- *      On failure: NULL ("data" is not a complete udp packet).
+ *      On failure: NULL ("data" is not a valid udp packet).  In this case,
+ *                  udpheader is bzero(3)-ed.
  */
 
 const char *
-socks_packet2string(const void *packet, int type);
+socks_packet2string(const void *packet, int isrequest);
 /*
- * debug function; dumps socks packet content
- * "packet" is a socks packet, "type" indicates it's type.
+ * debug function; dumps socks packet content.
+ * "packet" is a socks packet, "isrequest" is set if it is a request
+ * packet, false otherwise.
+ *
  * Returns:
  *      On success: 0
  *      On failure: -1
@@ -257,12 +259,12 @@ str2upper(char *string);
 
 char *
 socket2string(const int s, char *buf, size_t buflen)
-      __attribute__((__nonnull__(2)))
       __attribute__((__bounded__(__string__, 2, 3)));
 /*
  * Prints out address info for the socket "s".
  * "buf" gives the buffer to write the address info to, "buflen" the
- * size of "buf".
+ * size of "buf".  If buflen is zero, a statically allocated buffer will
+ * be used instead.
  *
  * Returns a pointer to buf.
  */
@@ -274,12 +276,24 @@ version2string(int version);
  */
 
 const char *
-atype2string(const int atype);
+atype2string(const atype_t atype);
 /*
  * Returns a printable representation of the atype "atype".
  */
 
-#if SOCKS_SERVER || BAREFOOTD
+char *
+routeoptions2string(const routeoptions_t *options, char *str, size_t strsize);
+/*
+ * Returns a printable representation of the atype "atype".
+ * If "str" and "strsize" is NULL and zero, the function returns a
+ * string that will be overwritten on the next call to this function.
+ */
+
+
+const char *loglevel2string(const int loglevel);
+/*
+ * returns a printable representation of the loglevel "loglevel".
+ */
 
 char *
 logtypes2string(const struct logtype_t *logtypes, char *str, size_t strsize)
@@ -293,6 +307,8 @@ logtypes2string(const struct logtype_t *logtypes, char *str, size_t strsize)
  * Returns: "str", NUL terminated.
  */
 
+#if !SOCKS_CLIENT
+
 char *
 timeouts2string(const struct timeout_t *timeouts, const char *prefix,
       char *str, size_t strsize)
@@ -417,4 +433,14 @@ privop2string(const priv_op_t op);
  * Returns a printable representation of "op".
  */
 
-#endif /* SOCKS_SERVER */
+#if COVENANT
+const char *httpcode2string(const int version, const int code);
+/*
+ * Returns a short printable representation of the http version
+ * "version" response code "code"
+ */
+
+#endif /* COVENANT */
+#endif /* !SOCKS_CLIENT */
+
+#endif /* !_TOSTRING_H_ */
