@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2004, 2005, 2008, 2009
+ * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2004, 2005, 2008, 2009, 2010,
+ *               2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -49,7 +50,7 @@
 #include "common.h"
 
 static const char rcsid[] =
-"$Id: Rgethostbyname.c,v 1.65 2009/10/23 11:43:34 karls Exp $";
+"$Id: Rgethostbyname.c,v 1.74 2011/05/18 13:48:45 karls Exp $";
 
 struct hostent *
 Rgethostbyname2(name, af)
@@ -61,6 +62,10 @@ Rgethostbyname2(name, af)
    static char *aliases[] = { NULL };
    struct in_addr ipindex;
    struct hostent *hostent;
+   char ipv4[sizeof(struct in_addr)];
+#if HAVE_IPV6_SUPPORT
+   char ipv6[sizeof(struct in6_addr)];
+#endif /* HAVE_IPV6_SUPPORT */
 
    clientinit();
 
@@ -104,29 +109,23 @@ Rgethostbyname2(name, af)
 
    if (hostent->h_addr_list == NULL) {
       /* * 2; NULL terminated and always only one valid entry (fake). */
-      if ((hostent->h_addr_list
-      = malloc(sizeof(hostent->h_addr_list) * 2)) == NULL)
+      if ((hostent->h_addr_list = malloc(sizeof(hostent->h_addr_list) * 2))
+      == NULL)
          return NULL;
       hostent->h_addr_list[1] = NULL;
    }
 
    switch (af) {
-      case AF_INET: {
-         static char ipv4[sizeof(struct in_addr)];
-
-         hostent->h_length     = sizeof(ipv4);
-         *hostent->h_addr_list = ipv4;
+      case AF_INET:
+         hostent->h_length       = sizeof(ipv4);
+         hostent->h_addr_list[0] = ipv4;
          break;
-      }
 
 #if HAVE_IPV6_SUPPORT
-      case AF_INET6: {
-         static char ipv6[struct in6_addr)];
-
-         hostent->h_length     = sizeof(ipv6);
-         *hostent->h_addr_list = ipv6;
+      case AF_INET6:
+         hostent->h_length        = sizeof(ipv6);
+         *hostent->h_addr_list[0] = ipv6;
          break;
-      }
 #endif /* HAVE_IPV6_SUPPORT */
 
       default:
@@ -197,7 +196,6 @@ Rgetaddrinfo(nodename, servname, hints, res)
       if (inet_pton(AF_INET6, nodename, addrbuf) == 1)
          fakeip_flag = 0;
 #endif /* HAVE_IPV6_SUPPORT */
-
    }
    else if (hints->ai_protocol == PF_INET) {
       if (inet_pton(AF_INET,  nodename, addrbuf) == 1)
@@ -210,6 +208,13 @@ Rgetaddrinfo(nodename, servname, hints, res)
          gaierr = getaddrinfo(nodename, servname, hints, res);
          if (gaierr == 0 || !fakeip_flag)
             return gaierr;
+
+         slog(LOG_DEBUG, "%s: getaddrinfo(%s, %s) failed: %s",
+         function,
+         nodename == NULL ? "null" : nodename,
+         servname == NULL ? "null" : servname,
+         gai_strerror(gaierr));
+
          break;
 
       case RESOLVEPROTOCOL_FAKE:
@@ -224,12 +229,6 @@ Rgetaddrinfo(nodename, servname, hints, res)
    if (!fakeip_flag || nodename == NULL)
       return EAI_NONAME;
 
-   if (sockscf.resolveprotocol != RESOLVEPROTOCOL_FAKE)
-      slog(LOG_DEBUG, "%s: getaddrinfo(%s, %s) failed: %s",
-      function,
-      nodename == NULL ? "null" : nodename,
-      servname == NULL ? "null" : servname, gai_strerror(gaierr));
-
    if ((ipindex.s_addr = socks_addfakeip(nodename)) == htonl(INADDR_NONE))
       return EAI_NONAME;
 
@@ -238,16 +237,18 @@ Rgetaddrinfo(nodename, servname, hints, res)
    SASSERTX(addrstr[sizeof(addrstr) - 1] == NUL);
 
    slog(LOG_DEBUG, "%s: faking ip address %s for (%s, %s)",
-   function, addrstr,
-   nodename == NULL ? "null" : nodename,
-   servname == NULL ? "null" : servname);
+        function,
+        addrstr,
+        nodename == NULL ? "null" : nodename,
+        servname == NULL ? "null" : servname);
 
    if (hints == NULL) {
       fakehints.ai_flags     = AI_NUMERICHOST;
       fakehints.ai_family    = PF_INET;
       fakehints.ai_socktype  = 0;
       fakehints.ai_protocol  = 0;
-   } else {
+   }
+   else {
       fakehints.ai_flags    = hints->ai_flags | AI_NUMERICHOST;
       fakehints.ai_family   = hints->ai_family;
       fakehints.ai_socktype = hints->ai_socktype;
@@ -283,7 +284,7 @@ Rgetipnodebyname2(name, af, flags, error_num)
    struct in_addr ipindex;
    struct hostent *hostent;
 
-   /* needs to be done before getipnodebyname calls. */
+   /* needs to be done before getipnodebyname() calls. */
    clientinit();
 
    slog(LOG_DEBUG, "%s: %s", function, name);
@@ -309,7 +310,8 @@ Rgetipnodebyname2(name, af, flags, error_num)
    if (h_errno != NO_RECOVERY)
       return hostent;
 
-   hostent = malloc(sizeof(struct hostent));
+   if ((hostent = malloc(sizeof(struct hostent))) == NULL)
+      return NULL;
 
    /* anything that fails from here is due to resource shortage. */
    h_errno = TRY_AGAIN;
@@ -366,7 +368,7 @@ Rgetipnodebyname2(name, af, flags, error_num)
       free(*addrlist);
       free(addrlist);
 
-        return NULL;
+      return NULL;
    }
 
    switch (af) {
@@ -390,9 +392,6 @@ Rgetipnodebyname2(name, af, flags, error_num)
          SERRX(af);
    }
 
-   slog(LOG_DEBUG, "%s: after inet_pton (0x%x, %s)",
-   function, (unsigned int)*addrlist, inet_ntoa(ipindex));
-
    hostent->h_addr_list = addrlist++;
    *addrlist = NULL;
 
