@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2005, 2008, 2009
+ * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2005, 2008, 2009, 2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -44,7 +44,7 @@
 #include "common.h"
 
 static const char rcsid[] =
-"$Id: authneg.c,v 1.91 2009/10/23 11:43:35 karls Exp $";
+"$Id: authneg.c,v 1.94 2011/06/19 14:33:57 michaels Exp $";
 
 int
 negotiate_method(s, packet, route)
@@ -67,7 +67,7 @@ negotiate_method(s, packet, route)
 
    if (sockscf.option.debug)
       slog(LOG_DEBUG, "%s: socket %d, %s",
-      function, s, socket2string(s, buf, sizeof(buf)));
+      function, s, socket2string(s, NULL, 0));
 
    SASSERTX(packet->gw.state.methodc > 0);
 
@@ -103,8 +103,8 @@ negotiate_method(s, packet, route)
 
    if ((rc = socks_recvfromn(s, response, sizeof(response), sizeof(response),
    0, NULL, NULL, packet->req.auth)) != sizeof(response)) {
-      swarn("%s: could not read server response for method to use, read %d/%ld",
-      function, rc, (long)sizeof(response));
+      swarn("%s: could not read server response for method to use, read %d/%lu",
+      function, rc, (unsigned long)sizeof(response));
 
       socks_blacklist(route);
 
