@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2008, 2009
+ * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2008, 2009, 2010
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -44,7 +44,7 @@
 #include "common.h"
 
 static const char rcsid[] =
-"$Id: serr.c,v 1.30 2009/10/23 11:43:37 karls Exp $";
+"$Id: serr.c,v 1.36 2011/06/13 09:55:00 michaels Exp $";
 
 #if SOCKS_CLIENT
 /* for errors, we want it logged. */
@@ -67,8 +67,9 @@ serr(int eval, const char *fmt, ...)
       va_end(ap);
 
       if (errno != 0)
-         bufused += snprintfn(&buf[bufused], sizeof(buf) - bufused,
-         ": %s (errno = %d)", strerror(errno), errno);
+         snprintf(&buf[bufused], sizeof(buf) - bufused,
+                  ": %s (errno = %d)",
+                  strerror(errno), errno);
 
       slog(LOG_ERR, "%s", buf);
    }
@@ -93,8 +94,8 @@ serrx(int eval, const char *fmt, ...)
       vslog(LOG_ERR, fmt, ap, apcopy);
 
       /* LINTED expression has null effect */
-      va_end(ap);
       va_end(apcopy);
+      va_end(ap);
    }
 
 #if SOCKS_CLIENT
@@ -119,10 +120,11 @@ swarn(const char *fmt, ...)
       bufused = vsnprintf(buf, sizeof(buf), fmt, ap);
 
       if (errno != 0)
-         bufused += snprintfn(&buf[bufused], sizeof(buf) - bufused,
-         ": %s (errno = %d)", strerror(errno), errno);
+         snprintf(&buf[bufused], sizeof(buf) - bufused,
+                  ": %s (errno = %d)",
+                  strerror(errno), errno);
 
-      slog(LOG_ERR, "%s", buf);
+      slog(LOG_WARNING, "%s", buf);
 
       /* LINTED expression has null effect */
       va_end(ap);
@@ -140,10 +142,10 @@ swarnx(const char *fmt, ...)
       va_start(ap, fmt);
       va_start(apcopy, fmt);
 
-      vslog(LOG_ERR, fmt, ap, apcopy);
+      vslog(LOG_WARNING, fmt, ap, apcopy);
 
       /* LINTED expression has null effect */
-      va_end(ap);
       va_end(apcopy);
+      va_end(ap);
    }
 }
