@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2004, 2005, 2008, 2009
+ * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2004, 2005, 2008, 2009, 2010,
+ *               2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -44,7 +45,7 @@
 #include "common.h"
 
 static const char rcsid[] =
-"$Id: udp.c,v 1.203 2009/10/23 10:11:45 karls Exp $";
+"$Id: udp.c,v 1.217 2011/08/01 12:23:39 michaels Exp $";
 
 /* ARGSUSED */
 ssize_t
@@ -57,18 +58,18 @@ Rsendto(s, msg, len, flags, to, tolen)
    socklen_t tolen;
 {
    const char *function = "Rsendto()";
-   struct socksfd_t socksfd;
-   struct sockshost_t host;
-   char srcstring[MAXSOCKADDRSTRING], dststring[sizeof(srcstring)];
-   void *nmsg;
+   struct socksfd_t socksfd, *p;
+   struct sockshost_t tohost;
    size_t nlen;
    ssize_t n;
+   char srcstring[MAXSOCKADDRSTRING], dststring[sizeof(srcstring)];
+   char nmsg[SOCKD_BUFSIZE];
 
    clientinit();
 
    slog(LOG_DEBUG, "%s: socket %d, len %lu, address %s",
    function, s, (long unsigned)len,
-   to == NULL ? "<none given>" : sockaddr2string(to, NULL, 0));
+   to == NULL ? "NULL" : sockaddr2string(to, NULL, 0));
 
    if (to != NULL && to->sa_family != AF_INET) {
       slog(LOG_DEBUG, "%s: unsupported address family '%d', system fallback",
@@ -83,17 +84,19 @@ Rsendto(s, msg, len, flags, to, tolen)
    }
 
    slog(LOG_DEBUG, "%s: route returned by udpsetup() is a %s route",
-   function,
-   proxyprotocols2string(&socksfd.route->gw.state.proxyprotocol, NULL, 0));
+                   function,
+                   proxyprotocols2string(&socksfd.route->gw.state.proxyprotocol,
+                   NULL, 0));
 
    if (socksfd.route->gw.state.proxyprotocol.direct) {
-      slog(LOG_DEBUG, "%s: using direct systemcalls for socket %d",
+      slog(LOG_DEBUG, "%s: using direct system calls for socket %d",
       function, s);
 
       return sendto(s, msg, len, flags, to, tolen);
    }
 
-   socksfd = *socks_getaddr(s, 1);
+   p = socks_getaddr(s, &socksfd, 1);
+   SASSERTX(p != NULL);
 
    if (socksfd.state.issyscall
    ||  socksfd.state.version == PROXY_UPNP)
@@ -109,7 +112,7 @@ Rsendto(s, msg, len, flags, to, tolen)
 
    if (to == NULL) {
       if (socksfd.state.udpconnect)
-         to = &socksfd.forus.connected;
+         tohost = socksfd.forus.connected;
       else { /* tcp. */
          n = socks_sendto(s, msg, len, flags, NULL, 0, &socksfd.state.auth);
 
@@ -122,31 +125,38 @@ Rsendto(s, msg, len, flags, to, tolen)
          return n;
       }
    }
+   else
+      fakesockaddr2sockshost(to, &tohost);
 
-   /* prefix a UDP header to the msg */
+
+   /*
+    * need to prefix a socks udp header to the message.  Copy the original
+    * payload into nmsg, which should have room to prefix the socks
+    * udpheader, and then send it.
+    */
+    slog(LOG_DEBUG, "%s: prefixing address %s to payload from client ...",
+    function, sockshost2string(&tohost, NULL, 0));
+
+   memcpy(nmsg, msg, len);
    nlen = len;
-   /* LINTED warning: cast discards 'const' from pointer target type */
-   if ((nmsg = udpheader_add(fakesockaddr2sockshost(to, &host), msg, &nlen,
-   len)) == NULL) {
-      errno = ENOBUFS;
+   if (udpheader_add(&tohost, nmsg, &nlen, sizeof(nmsg)) == NULL)
       return -1;
-   }
 
    n = socks_sendto(s, nmsg, nlen, flags,
-   socksfd.state.udpconnect ? NULL : &socksfd.reply,
-   socksfd.state.udpconnect ? (socklen_t)0 : sizeof(socksfd.reply),
-   &socksfd.state.auth);
+                    socksfd.state.udpconnect ? NULL : &socksfd.reply,
+                    socksfd.state.udpconnect ?
+                        (socklen_t)0 : sizeof(socksfd.reply),
+                    &socksfd.state.auth);
 
    n -= (ssize_t)(nlen - len);
 
-   if (msg != nmsg)
-      free(nmsg);
-
    slog(LOG_DEBUG, "%s: %s: %s -> %s (%lu)",
-   function, protocol2string(SOCKS_UDP),
-   sockaddr2string(&socksfd.local, dststring, sizeof(dststring)),
-   sockaddr2string(&socksfd.reply, srcstring, sizeof(srcstring)),
-   (unsigned long)n);
+                   function, protocol2string(SOCKS_UDP),
+                   sockaddr2string(&socksfd.local, dststring,
+                                   sizeof(dststring)),
+                   sockaddr2string(&socksfd.reply, srcstring,
+                                   sizeof(srcstring)),
+                   (unsigned long)n);
 
    return MAX(-1, n);
 }
@@ -163,6 +173,7 @@ Rrecvfrom(s, buf, len, flags, from, fromlen)
    const char *function = "Rrecvfrom()";
    struct socksfd_t socksfd;
    struct udpheader_t header;
+   struct route_t *route;
    struct sockaddr newfrom;
    socklen_t newfromlen;
    char srcstring[MAXSOCKADDRSTRING], dststring[sizeof(srcstring)], *newbuf;
@@ -171,25 +182,23 @@ Rrecvfrom(s, buf, len, flags, from, fromlen)
 
    slog(LOG_DEBUG, "%s: socket %d, len %lu", function, s, (long unsigned)len);
 
-   if (!socks_addrisours(s, 1)) {
+   if (!socks_addrisours(s, &socksfd, 1)) {
       socks_rmaddr(s, 1);
       return recvfrom(s, buf, len, flags, from, fromlen);
    }
 
-   if ((socksfd.route = udpsetup(s, from, SOCKS_RECV)) == NULL) {
+   if ((route = udpsetup(s, from, SOCKS_RECV)) == NULL) {
       slog(LOG_DEBUG, "%s: udpsetup() failed for socket %d", function, s);
       return -1;
    }
 
-   if (socksfd.route->gw.state.proxyprotocol.direct) {
+   if (route->gw.state.proxyprotocol.direct) {
       slog(LOG_DEBUG, "%s: using direct system calls for socket %d",
       function, s);
 
       return recvfrom(s, buf, len, flags, from, fromlen);
    }
 
-   socksfd = *socks_getaddr(s, 1);
-
    if (socksfd.state.issyscall
    ||  socksfd.state.version == PROXY_UPNP)
       return recvfrom(s, buf, len, flags, from, fromlen);
@@ -203,7 +212,7 @@ Rrecvfrom(s, buf, len, flags, from, fromlen)
    }
 
    if (socksfd.state.protocol.tcp) {
-      const struct sockaddr *forus;
+      const struct sockshost_t *forus;
 
       if (socksfd.state.err != 0) {
          errno = socksfd.state.err;
@@ -227,7 +236,7 @@ Rrecvfrom(s, buf, len, flags, from, fromlen)
          case SOCKS_BIND:
             forus = &socksfd.forus.accepted;
 
-            if (forus->sa_family == 0) {
+            if (forus->atype == (unsigned char)SOCKS_ADDR_NOTSET) {
                swarnx("%s: strange ... trying to read from socket %d, "
                       "which is for bind, but no bind-reply received yet ...",
                       function, s);
@@ -240,11 +249,13 @@ Rrecvfrom(s, buf, len, flags, from, fromlen)
       }
 
       slog(LOG_DEBUG, "%s: %s: %s -> %s (%ld: %s)",
-      function, protocol2string(SOCKS_TCP),
-      forus == NULL ?
-      "<NULL>" : sockaddr2string(forus, srcstring, sizeof(srcstring)),
-      sockaddr2string(&socksfd.local, dststring, sizeof(dststring)),
-      (long)n, strerror(errno));
+           function,
+           protocol2string(SOCKS_TCP),
+           forus == NULL ?
+               "<NULL>" : sockshost2string(forus, srcstring, sizeof(srcstring)),
+           sockaddr2string(&socksfd.local, dststring, sizeof(dststring)),
+           (long)n,
+           strerror(errno));
 
       return n;
    }
@@ -292,10 +303,10 @@ Rrecvfrom(s, buf, len, flags, from, fromlen)
    free(newbuf);
 
    slog(LOG_DEBUG, "%s: %s: %s -> %s (%ld)",
-   function, protocol2string(SOCKS_UDP),
-   sockaddr2string(&newfrom, srcstring, sizeof(srcstring)),
-   sockaddr2string(&socksfd.local, dststring, sizeof(dststring)),
-   (long)n);
+        function, protocol2string(SOCKS_UDP),
+        sockaddr2string(&newfrom, srcstring, sizeof(srcstring)),
+        sockaddr2string(&socksfd.local, dststring, sizeof(dststring)),
+        (long)n);
 
    if (from != NULL) {
       *fromlen = MIN(*fromlen, newfromlen);
@@ -313,7 +324,6 @@ udpsetup(s, to, type)
 {
    const char *function = "udpsetup()";
    static struct route_t directroute;
-   const struct socksfd_t *socksfdptr;
    struct socksfd_t socksfd;
    struct authmethod_t auth;
    struct socks_t packet;
@@ -344,26 +354,24 @@ udpsetup(s, to, type)
 
       default:
          slog(LOG_DEBUG, "%s: unsupported af %d",
-         function, socksfd.local.sa_family);
+              function, socksfd.local.sa_family);
 
          return &directroute;
    }
    sockaddr2sockshost(&socksfd.local, &src);
 
-
    slog(LOG_DEBUG, "%s: socket %d, type = %s",
    function, s, type == SOCKS_RECV ? "receive" : "send");
 
-   if (!socks_addrisours(s, 1))
-      socks_rmaddr(s, 1);
-
-   if ((socksfdptr = socks_getaddr(s, 1)) != NULL) {
+   if (socks_addrisours(s, &socksfd, 1)) {
       slog(LOG_DEBUG, "%s: route already setup for socket %d", function, s);
-      return socksfdptr->route; /* all set up. */
+      return socksfd.route;
    }
 
+   socks_rmaddr(s, 1);
+
    if (socks_socketisforlan(s)) {
-      slog(LOG_DEBUG, "%s: socket %d is for lan only, system fallback",
+      slog(LOG_DEBUG, "%s: socket %d is for lan only.  System fallback",
       function, s);
 
       return &directroute;
@@ -378,7 +386,7 @@ udpsetup(s, to, type)
           * should already have been bound, so socks_addrisours()
           * should have been true.
           */
-         swarnx("%s: receive on udp socket not previously sent on is "
+         swarnx("%s: receive on a udp socket not previously sent on is "
                 "not supported by the socks protocol, returning direct route",
                 function);
 
@@ -396,7 +404,7 @@ udpsetup(s, to, type)
 
                len = sizeof(val);
                if (getsockopt(s, SOL_SOCKET, SO_TYPE, &val, &len) != 0) {
-                  slog(LOG_DEBUG, "%s: getsockopt(SO_TYPE): %s",
+                  slog(LOG_DEBUG, "%s: getsockopt(SO_TYPE) failed: %s",
                   function, strerror(errno));
 
                   return &directroute;
@@ -411,6 +419,7 @@ udpsetup(s, to, type)
                           "%s: socket %d is unknown, but has a stream "
                           "peer (%s), returning direct route",
                           function, s, sockaddr2string(&addr, NULL, 0));
+
                      return &directroute;
 
                   default:
@@ -451,8 +460,8 @@ udpsetup(s, to, type)
    packet.req.version   = packet.version;
    packet.req.command   = SOCKS_UDPASSOCIATE;
 #if 0 /*
-       * some (nec-based) socks-server misinterpret this to mean something
-       * completely different.
+       * some (nec-based) socks-servers misinterpret this to mean something
+       * completely different than what the draft says.
        */
    packet.req.flag     |= SOCKS_USECLIENTPORT;
 #endif
@@ -481,6 +490,9 @@ udpsetup(s, to, type)
             swarn("%s: failed to create control socket", function);
             return NULL;
          }
+
+         slog(LOG_DEBUG, "%s: control socket %d created for data socket %d",
+              function, socksfd.control, s);
          break;
 
       default:
@@ -521,14 +533,15 @@ udpsetup(s, to, type)
          close(socksfd.control);
          return NULL;
       }
+
+      if (getsockname(s, &socksfd.local, &len) != 0) {
+         swarn("%s: getsockname(s)", function);
+
+         close(socksfd.control);
+         return NULL;
+      }
    }
 
-   if (getsockname(s, &socksfd.local, &len) != 0) {
-      swarn("%s: getsockname(s)", function);
-
-      close(socksfd.control);
-      return NULL;
-   }
    sockaddr2sockshost(&socksfd.local, &packet.req.host);
 
    if (socks_negotiate(s, socksfd.control, &packet, socksfd.route) != 0) {
@@ -556,7 +569,7 @@ udpsetup(s, to, type)
 
    if (shouldconnect) {
       socksfd.state.udpconnect = 1;
-      socksfd.forus.connected  = *to;
+      sockaddr2sockshost(to, &socksfd.forus.connected);
    }
 
    if (socksfd.state.version == PROXY_UPNP) {
