@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008, 2009, 2010
+ * Copyright (c) 2008, 2009, 2010, 2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,17 +42,11 @@
  */
 
 static const char rcsid[] =
-"$Id: upnp.c,v 1.62.2.2 2010/05/24 16:38:36 karls Exp $";
+"$Id: upnp.c,v 1.85 2011/07/21 14:09:19 karls Exp $";
 
 #include "common.h"
 
-#if HAVE_LIBMINIUPNP
-#include <miniupnpc/miniupnpc.h>
-#include <miniupnpc/upnpcommands.h>
-#include <miniupnpc/upnperrors.h>
-#else
 #include "upnp.h"
-#endif /* HAVE_LIBMINIUPNP */
 
 #if HAVE_LIBMINIUPNP
 #if SOCKS_CLIENT
@@ -75,11 +69,11 @@ socks_initupnp(gw, state)
    struct UPNPUrls url;
    struct IGDdatas data;
    char myaddr[INET_ADDRSTRLEN], gwstring[MAXGWSTRING],
-        addrstring[MAXSOCKSHOSTSTRING];
+        addrstring[MAXSOCKSHOSTSTRING], emsg[256];
    int devtype, rc;
 #endif /* HAVE_LIBMINIUPNP */
 
-   slog(LOG_DEBUG, function);
+   slog(LOG_DEBUG, "%s", function);
 
 #if !HAVE_LIBMINIUPNP
    return -1;
@@ -87,6 +81,9 @@ socks_initupnp(gw, state)
    if (*state->upnp.controlurl != NUL)
       return 0;
 
+   bzero(&url, sizeof(url));
+   bzero(&data, sizeof(data));
+
    if (gw->atype == SOCKS_ADDR_URL) {
       slog(LOG_DEBUG, "%s: using IGD at \"%s\"\n", function, gw->addr.urlname);
 
@@ -107,7 +104,8 @@ socks_initupnp(gw, state)
       struct UPNPDev *p;
 
       gwaddr2sockshost(gw, &host);
-      SASSERTX(host.atype == SOCKS_ADDR_IPV4);
+
+      SASSERTX(host.atype == (unsigned char)SOCKS_ADDR_IPV4);
       inet_ntop(AF_INET, &host.addr.ipv4, addrstring, sizeof(addrstring));
 
       slog(LOG_DEBUG, "%s: doing upnp discovery on interface of addr %s (%s)",
@@ -130,8 +128,13 @@ socks_initupnp(gw, state)
          struct sockaddr saddr;
          struct sockaddr_in smask;
 
-         if (urlstring2sockaddr(p->descURL, &saddr) == NULL)
+         if (urlstring2sockaddr(p->descURL, &saddr, emsg, sizeof(emsg))
+         == NULL) {
+            slog(LOG_DEBUG, "%s: urlstring2sockaddr() failed: %s", 
+                 function, emsg);
+
             continue;
+         }
 
          bzero(&smask, sizeof(smask));
          smask.sin_family      = AF_INET;
@@ -172,23 +175,35 @@ socks_initupnp(gw, state)
                errno = ENETUNREACH;
 
             rc = -1;
+            break;
 
          default:
             swarnx("%s: unknown return code from UPNP_GetValidIGD(): %d",
             function, devtype);
-            rc = -1;
 
             if (errno == 0)
                errno = ENETUNREACH;
+
+            rc = -1;
       }
 
       freeUPNPDevlist(dev);
    }
 
-   SASSERTX(strlen(url.controlURL) < sizeof(state->upnp.controlurl));
-   strcpy(state->upnp.controlurl, url.controlURL);
-   SASSERTX(strlen(data.servicetype) < sizeof(state->upnp.servicetype));
-   strcpy(state->upnp.servicetype, data.servicetype);
+   if (rc == 0) {
+      SASSERTX(strlen(url.controlURL) < sizeof(state->upnp.controlurl));
+      strcpy(state->upnp.controlurl, url.controlURL);
+
+#if HAVE_LIBMINIUPNP13
+      SASSERTX(strlen(data.servicetype) < sizeof(state->upnp.servicetype));
+      strcpy(state->upnp.servicetype, data.servicetype);
+#elif HAVE_LIBMINIUPNP14
+      SASSERTX(strlen(data.CIF.servicetype) < sizeof(state->upnp.servicetype));
+      strcpy(state->upnp.servicetype, data.CIF.servicetype);
+#else
+#  error "unexpected miniupnp version"
+#endif /* HAVE_LIBMINIUPNP14 */
+   }
 
    FreeUPNPUrls(&url);
 
@@ -206,11 +221,11 @@ upnp_negotiate(s, packet, state)
 #if HAVE_LIBMINIUPNP
    struct sockaddr_in addr;
    socklen_t addrlen;
-   char straddr[INET_ADDRSTRLEN], strport[sizeof("65535")];
+   char straddr[INET_ADDRSTRLEN], strport[sizeof("65535")], emsg[256];
    int rc;
 #endif /* HAVE_LIBMINIUPNP */
 
-   slog(LOG_DEBUG, function);
+   slog(LOG_DEBUG, "%s", function);
 
 #if !HAVE_LIBMINIUPNP
    SERRX(0);
@@ -218,7 +233,7 @@ upnp_negotiate(s, packet, state)
    packet->res.version = PROXY_UPNP;
 
    switch (packet->req.command) {
-      case SOCKS_CONNECT:
+      case SOCKS_CONNECT: {
          /*
           * Can only find out what the external ip address of the device is.
           *
@@ -232,17 +247,25 @@ upnp_negotiate(s, packet, state)
           * to the socks client.
           */
 
-         if (socks_connecthost(s, &packet->req.host) != 0)
-            if (!ERRNOISINPROGRESS(errno)) {
+         if (socks_connecthost(s,
+                               &packet->req.host,
+                               NULL,
+                               sockscf.timeout.connect
+                                 ? (long)sockscf.timeout.connect : -1,
+                               emsg,
+                               sizeof(emsg)) != 0)
+            if (errno == EINPROGRESS) {
                slog(LOG_DEBUG, "%s: socks_connecthost(%s) failed: %s",
-               function, sockshost2string(&packet->req.host, NULL, 0),
-               strerror(errno));
+                    function,
+                    sockshost2string(&packet->req.host, NULL, 0),
+                    emsg);
 
-               packet->res.reply = UPNP_FAILURE;
+               socks_set_responsevalue(&packet->res, UPNP_FAILURE);
                return -1;
             }
 
          /* FALLTHROUGH */
+      }
 
       case SOCKS_UDPASSOCIATE: {
          /*
@@ -256,7 +279,7 @@ upnp_negotiate(s, packet, state)
           */
          const int errno_s = errno;
 
-         packet->res.host.atype              = SOCKS_ADDR_IPV4;
+         packet->res.host.atype = (unsigned char)SOCKS_ADDR_IPV4;
 
 #if SOCKS_CLIENT
          /* will be filled with real value if user ever does getsockname(2). */
@@ -265,7 +288,7 @@ upnp_negotiate(s, packet, state)
          if ((rc = UPNP_GetExternalIPAddress(state->upnp.controlurl,
          state->upnp.servicetype, straddr)) != UPNPCOMMAND_SUCCESS) {
             swarnx("failed to get external ip address of upnp device: %d", rc);
-            packet->res.reply = UPNP_FAILURE;
+            socks_set_responsevalue(&packet->res, UPNP_FAILURE);
 
             return -1;
          }
@@ -282,7 +305,7 @@ upnp_negotiate(s, packet, state)
          }
          packet->res.host.port = TOIN(&addr)->sin_port;
          slog(LOG_DEBUG, "%s: will never know for sure, but hoping IGD "
-                         "will use same port as we (%d)",
+                         "will use same port as we; %d",
                          function, ntohs(packet->res.host.port));
 
          rc    = 0;
@@ -307,14 +330,14 @@ upnp_negotiate(s, packet, state)
          addrlen = sizeof(addr);
          if (getsockname(s, (struct sockaddr *)&addr, &addrlen) != 0) {
             swarn("getsockname()");
-            packet->res.reply = UPNP_FAILURE;
+            socks_set_responsevalue(&packet->res, UPNP_FAILURE);
             return -1;
          }
 
          if ((rc = UPNP_GetExternalIPAddress(state->upnp.controlurl,
          state->upnp.servicetype, straddr)) != UPNPCOMMAND_SUCCESS) {
             swarnx("failed to get external ip address of upnp device: %d", rc);
-            packet->res.reply = UPNP_FAILURE;
+            socks_set_responsevalue(&packet->res, UPNP_FAILURE);
             return -1;
          }
          else {
@@ -341,11 +364,11 @@ upnp_negotiate(s, packet, state)
                    if (ifname2sockaddr(packet->gw.addr.addr.ifname, 0, &tmpaddr,
                    NULL) == NULL) {
                      swarn("ifname2sockaddr(%s)", packet->gw.addr.addr.ifname);
-                     packet->res.reply = UPNP_FAILURE;
+                     socks_set_responsevalue(&packet->res, UPNP_FAILURE);
                      return -1;
                   }
 
-                  /* just want the ipaddr.  Port number etc. remains the same. */
+                  /* just want the ipaddr.  Port number, etc., remains same. */
                   addr.sin_addr = TOIN(&tmpaddr)->sin_addr;
                   break;
                }
@@ -354,9 +377,16 @@ upnp_negotiate(s, packet, state)
                   socklen_t tmpaddrlen;
                   int ss;
 
-                  if (urlstring2sockaddr(packet->gw.addr.addr.urlname, &tmpaddr)
+                  if (urlstring2sockaddr(packet->gw.addr.addr.urlname,
+                                         &tmpaddr,
+                                         emsg,
+                                         sizeof(emsg))
                   == NULL) {
-                     packet->res.reply = UPNP_FAILURE;
+                     slog(LOG_DEBUG, "%s: urlstring2sockaddr() failed: %s", 
+                          function, emsg);
+
+                     socks_set_responsevalue(&packet->res, UPNP_FAILURE);
+
                      return -1;
                   }
 
@@ -371,23 +401,28 @@ upnp_negotiate(s, packet, state)
 
                   if ((ss = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
                      swarn("%s: socket()", function);
-                     packet->res.reply = UPNP_FAILURE;
+                     socks_set_responsevalue(&packet->res, UPNP_FAILURE);
+
                      return -1;
                   }
 
                   if (connect(ss, &tmpaddr, sizeof(tmpaddr)) != 0) {
                      swarn("%s: connect(%s)",
                      function, sockaddr2string(&tmpaddr, NULL, 0));
-                     packet->res.reply = UPNP_FAILURE;
+
+                     socks_set_responsevalue(&packet->res, UPNP_FAILURE);
                      close(ss);
+
                      return -1;
                   }
 
                   tmpaddrlen = sizeof(tmpaddr);
                   if (getsockname(ss, &tmpaddr, &tmpaddrlen) != 0) {
                      swarn("%s: getsockname()", function);
-                     packet->res.reply = UPNP_FAILURE;
+
+                     socks_set_responsevalue(&packet->res, UPNP_FAILURE);
                      close(ss);
+
                      return -1;
                   }
 
@@ -405,14 +440,14 @@ upnp_negotiate(s, packet, state)
          if (inet_ntop(addr.sin_family, &addr.sin_addr, straddr,
          sizeof(straddr)) == NULL) {
             swarn("inet_ntop()");
-            packet->res.reply = UPNP_FAILURE;
+            socks_set_responsevalue(&packet->res, UPNP_FAILURE);
             return -1;
          }
 
          len = sizeof(val);
          if (getsockopt(s, SOL_SOCKET, SO_TYPE, &val, &len) != 0) {
             swarn("getsockopt()");
-            packet->res.reply = UPNP_FAILURE;
+            socks_set_responsevalue(&packet->res, UPNP_FAILURE);
             return -1;
          }
          switch (val) {
@@ -426,7 +461,7 @@ upnp_negotiate(s, packet, state)
 
             default:
                swarn("unknown protocol type %d", val);
-               packet->res.reply = UPNP_FAILURE;
+               socks_set_responsevalue(&packet->res, UPNP_FAILURE);
                return -1;
          }
 
@@ -446,7 +481,7 @@ upnp_negotiate(s, packet, state)
                swarnx("%s: UPNP_AddPortMapping() failed: %s",
                function, strupnperror(rc));
 
-               packet->res.reply = UPNP_FAILURE;
+               socks_set_responsevalue(&packet->res, UPNP_FAILURE);
                return -1;
          }
          else
@@ -491,7 +526,7 @@ upnp_negotiate(s, packet, state)
          SERRX(packet->req.command);
    }
 
-   packet->res.reply = UPNP_SUCCESS;
+   socks_set_responsevalue(&packet->res, UPNP_SUCCESS);
 
    return 0;
 #endif /* !HAVE_LIBMINIUPNP */
@@ -524,7 +559,7 @@ upnpcleanup(s)
    const int s;
 {
    const char *function = "upnpcleanup()";
-   struct socksfd_t *socksfd;
+   struct socksfd_t socksfd;
    int rc, current, last;
 
    slog(LOG_DEBUG, "%s: socket %d", function, s);
@@ -545,33 +580,33 @@ upnpcleanup(s)
       if (deleting == current)
          continue;
 
-      if ((socksfd = socks_getaddr(current, 0 /* XXX */)) == NULL)
+      if (socks_getaddr(current, &socksfd, 0 /* XXX */) == NULL)
          continue;
 
-      if (socksfd->state.version != PROXY_UPNP)
+      if (socksfd.state.version != PROXY_UPNP)
          continue;
 
       slog(LOG_DEBUG, "%s: socket %d has upnp session set up for command "
                       "%s, accept pending: %d",
-                      function, current, command2string(socksfd->state.command),
-                      socksfd->state.acceptpending);
+                      function, current, command2string(socksfd.state.command),
+                      socksfd.state.acceptpending);
 
-      if (socksfd->state.command != SOCKS_BIND)
+      if (socksfd.state.command != SOCKS_BIND)
          continue;
 
       /*
        * Is this the socket we listened on?  Or just a client we accept(2)'ed?
        * The port mapping is just created for the first case.
        */
-      if (!socksfd->state.acceptpending)
+      if (!socksfd.state.acceptpending)
          continue; /* just a client we accepted. */
 
       snprintf(port, sizeof(port), "%d",
-      ntohs(TOCIN(&socksfd->remote)->sin_port));
+      ntohs(TOCIN(&socksfd.remote)->sin_port));
 
-      if (socksfd->state.protocol.tcp)
+      if (socksfd.state.protocol.tcp)
          snprintf(protocol, sizeof(protocol), PROTOCOL_TCPs);
-      else if (socksfd->state.protocol.udp)
+      else if (socksfd.state.protocol.udp)
          snprintf(protocol, sizeof(protocol), PROTOCOL_UDPs);
       else {
          SWARNX(0);
@@ -591,8 +626,8 @@ upnpcleanup(s)
       deleting = current;
 
       if ((rc
-      = UPNP_DeletePortMapping(socksfd->route->gw.state.data.upnp.controlurl,
-                               socksfd->route->gw.state.data.upnp.servicetype,
+      = UPNP_DeletePortMapping(socksfd.route->gw.state.data.upnp.controlurl,
+                               socksfd.route->gw.state.data.upnp.servicetype,
                                port, protocol, NULL)) != UPNPCOMMAND_SUCCESS)
             swarnx("%s: UPNP_DeletePortMapping(%s, %s) failed: %s",
             function, port, protocol, strupnperror(rc));
@@ -618,6 +653,6 @@ void
 upnpcleanup(s)
    const int s;
 {
-      return;
+        return;
 }
 #endif /* !HAVE_LIBMINIUPNP */
