@@ -1,31 +1,31 @@
-/* $Id: getifaddrs.c,v 1.12.2.1 2010/05/04 10:56:16 karls Exp $ */
+/* $Id: getifaddrs.c,v 1.17 2011/05/18 13:11:07 karls Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "autoconf.h"
 #endif /* HAVE_CONFIG_H */
 
-#include "common.h"
+#include "osdep.h"
 
 /*
  * Copyright (c) 2000 - 2002, 2005 Kungliga Tekniska Högskolan
  * (Royal Institute of Technology, Stockholm, Sweden).
  * All rights reserved.
- * 
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
- * 
+ *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
- * 
+ *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 
+ *
  * 3. Neither the name of the Institute nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
- * 
+ *
  * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
@@ -42,7 +42,7 @@
 #if 0
 #ifdef HAVE_CONFIG_H
 #include <config.h>
-RCSID("$Id: getifaddrs.c,v 1.12.2.1 2010/05/04 10:56:16 karls Exp $");
+RCSID("$Id: getifaddrs.c,v 1.17 2011/05/18 13:11:07 karls Exp $");
 #endif
 #include "roken.h"
 #endif
@@ -107,7 +107,7 @@ struct mbuf;
  * 3. Neither the name of the author nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
- * 
+ *
  * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
@@ -129,10 +129,10 @@ struct mbuf;
 #include <malloc.h>
 #include <errno.h>
 #include <unistd.h>
+#endif
 
 #include <sys/socket.h>
 #include <asm/types.h>
-#endif
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
 #include <sys/types.h>
@@ -145,7 +145,7 @@ struct mbuf;
 #include <net/if_arp.h>
 #include <netinet/in.h>
 
-#include "ifaddrs.h_compat"
+#include "ifaddrs_compat.h"
 
 #define __set_errno(e) (errno = (e))
 #define __close(fd) (close(fd))
@@ -208,9 +208,9 @@ ifa_sa_len(sa_family_t family, int len)
   return size;
 }
 
-static void 
-ifa_make_sockaddr(sa_family_t family, 
-		  struct sockaddr *sa, 
+static void
+ifa_make_sockaddr(sa_family_t family,
+		  struct sockaddr *sa,
 		  void *p, size_t len,
 		  uint32_t scope, uint32_t scopeid)
 {
@@ -242,8 +242,8 @@ ifa_make_sockaddr(sa_family_t family,
 
 #ifndef IFA_NETMASK
 static struct sockaddr *
-ifa_make_sockaddr_mask(sa_family_t family, 
-		       struct sockaddr *sa, 
+ifa_make_sockaddr_mask(sa_family_t family,
+		       struct sockaddr *sa,
 		       uint32_t prefixlen)
 {
   int i;
@@ -289,7 +289,7 @@ ifa_make_sockaddr_mask(sa_family_t family,
 #endif
 
 /* ====================================================================== */
-static int 
+static int
 nl_sendreq(int sd, int request, int flags, int *seq)
 {
   char reqbuf[NLMSG_ALIGN(sizeof(struct nlmsghdr)) +
@@ -315,9 +315,9 @@ nl_sendreq(int sd, int request, int flags, int *seq)
 		 (struct sockaddr *)&nladdr, sizeof(nladdr)));
 }
 
-static int 
-nl_recvmsg(int sd, int request, int seq, 
-	   void *buf, size_t buflen, 
+static int
+nl_recvmsg(int sd, int request, int seq,
+	   void *buf, size_t buflen,
 	   int *flags)
 {
   struct msghdr msg;
@@ -342,8 +342,8 @@ nl_recvmsg(int sd, int request, int seq,
   return read_len;
 }
 
-static int 
-nl_getmsg(int sd, int request, int seq, 
+static int
+nl_getmsg(int sd, int request, int seq,
 	  struct nlmsghdr **nlhp,
 	  int *done)
 {
@@ -462,7 +462,7 @@ nl_getlist(int sd, int seq,
 }
 
 /* ---------------------------------------------------------------------- */
-static void 
+static void
 free_nlmsglist(struct nlmsg_list *nlm0)
 {
   struct nlmsg_list *nlm, *nlm_next;
@@ -479,7 +479,7 @@ free_nlmsglist(struct nlmsg_list *nlm0)
   __set_errno(saved_errno);
 }
 
-static void 
+static void
 free_data(void *data, void *ifdata)
 {
   int saved_errno = errno;
@@ -489,7 +489,7 @@ free_data(void *data, void *ifdata)
 }
 
 /* ---------------------------------------------------------------------- */
-static void 
+static void
 nl_close(int sd)
 {
   int saved_errno = errno;
@@ -498,7 +498,7 @@ nl_close(int sd)
 }
 
 /* ---------------------------------------------------------------------- */
-static int 
+static int
 nl_open(void)
 {
   struct sockaddr_nl nladdr;
@@ -576,7 +576,7 @@ getifaddrs(struct ifaddrs **ifap)
 		    NLMSG_ALIGN(sizeof(struct ifaddrs[icnt]))
 		    + dlen + xlen + nlen);
       ifa = (struct ifaddrs *)data;
-      ifdata = calloc(1, 
+      ifdata = calloc(1,
 		      NLMSG_ALIGN(sizeof(char *[max_ifindex+1]))
 		      + NLMSG_ALIGN(sizeof(uint16_t [max_ifindex+1])));
       if (ifap != NULL)
@@ -603,8 +603,8 @@ getifaddrs(struct ifaddrs **ifap)
       int nlmlen = nlm->size;
       if (!(nlh0 = nlm->nlh))
 	continue;
-      for (nlh = nlh0; 
-	   NLMSG_OK(nlh, nlmlen); 
+      for (nlh = nlh0;
+	   NLMSG_OK(nlh, nlmlen);
 	   nlh=NLMSG_NEXT(nlh,nlmlen)){
 	struct ifinfomsg *ifim = NULL;
 	struct ifaddrmsg *ifam = NULL;
@@ -650,7 +650,7 @@ getifaddrs(struct ifaddrs **ifap)
 	default:
 	  continue;
 	}
-	
+
 	if (!build){
 	  if (max_ifindex < nlm_index)
 	    max_ifindex = nlm_index;
@@ -841,8 +841,8 @@ getifaddrs(struct ifaddrs **ifap)
 	  if (ifa->ifa_name == NULL)
 	    ifa->ifa_name = iflist[nlm_index];
 #ifndef IFA_NETMASK
-	  if (ifa->ifa_addr && 
-	      ifa->ifa_addr->sa_family != AF_UNSPEC && 
+	  if (ifa->ifa_addr &&
+	      ifa->ifa_addr->sa_family != AF_UNSPEC &&
 	      ifa->ifa_addr->sa_family != AF_PACKET){
 	    ifa->ifa_netmask = (struct sockaddr *)data;
 	    ifa_make_sockaddr_mask(ifa->ifa_addr->sa_family, ifa->ifa_netmask, nlm_prefixlen);
@@ -878,7 +878,7 @@ getifaddrs(struct ifaddrs **ifap)
  */
 
 static int
-getifaddrs2(struct ifaddrs **ifap, 
+getifaddrs2(struct ifaddrs **ifap,
 	    int af, int siocgifconf, int siocgifflags,
 	    size_t ifreq_sz)
 {
@@ -1000,7 +1000,7 @@ getifaddrs2(struct ifaddrs **ifap,
 		ret = ENOMEM;
 		goto error_out;
 	    }
-	    memcpy((*end)->ifa_broadaddr, &ifr->ifr_broadaddr, 
+	    memcpy((*end)->ifa_broadaddr, &ifr->ifr_broadaddr,
 		   sizeof(ifr->ifr_broadaddr));
 	} else if (ifreq.ifr_flags & IFF_POINTOPOINT) {
 	    (*end)->ifa_dstaddr = malloc(sizeof(ifr->ifr_dstaddr));
@@ -1008,7 +1008,7 @@ getifaddrs2(struct ifaddrs **ifap,
 		ret = ENOMEM;
 		goto error_out;
 	    }
-	    memcpy((*end)->ifa_dstaddr, &ifr->ifr_dstaddr, 
+	    memcpy((*end)->ifa_dstaddr, &ifr->ifr_dstaddr,
 		   sizeof(ifr->ifr_dstaddr));
 	} else
 	    (*end)->ifa_dstaddr = NULL;
@@ -1019,7 +1019,6 @@ getifaddrs2(struct ifaddrs **ifap,
 	(*end)->ifa_data = NULL;
 
 	end = &(*end)->ifa_next;
-	
     }
     *ifap = start;
     close(fd);
@@ -1030,14 +1029,13 @@ getifaddrs2(struct ifaddrs **ifap,
     close(fd);
     free(buf);
     errno = ret;
-    slog(LOG_DEBUG, "getifaddrs2: failed %s", strerror(errno));
     return -1;
 }
 
 #if 0
 #if defined(HAVE_IPV6) && defined(SIOCGLIFCONF) && defined(SIOCGLIFFLAGS)
 static int
-getlifaddrs2(struct ifaddrs **ifap, 
+getlifaddrs2(struct ifaddrs **ifap,
 	     int af, int siocgifconf, int siocgifflags,
 	     size_t ifreq_sz)
 {
@@ -1164,7 +1162,7 @@ getlifaddrs2(struct ifaddrs **ifap,
 		ret = ENOMEM;
 		goto error_out;
 	    }
-	    memcpy((*end)->ifa_broadaddr, &ifr->ifr_broadaddr, 
+	    memcpy((*end)->ifa_broadaddr, &ifr->ifr_broadaddr,
 		   sizeof(ifr->ifr_broadaddr));
 	} else if (ifreq.ifr_flags & IFF_POINTOPOINT) {
 	    (*end)->ifa_dstaddr = malloc(sizeof(ifr->ifr_dstaddr));
@@ -1172,7 +1170,7 @@ getlifaddrs2(struct ifaddrs **ifap,
 		ret = ENOMEM;
 		goto error_out;
 	    }
-	    memcpy((*end)->ifa_dstaddr, &ifr->ifr_dstaddr, 
+	    memcpy((*end)->ifa_dstaddr, &ifr->ifr_dstaddr,
 		   sizeof(ifr->ifr_dstaddr));
 	} else
 	    (*end)->ifa_dstaddr = NULL;
@@ -1193,7 +1191,6 @@ getlifaddrs2(struct ifaddrs **ifap,
     close(fd);
     free(buf);
     errno = ret;
-    slog(LOG_DEBUG, "getlifaddrs2: failed %s", strerror(errno));
     return -1;
 }
 #endif /* defined(HAVE_IPV6) && defined(SIOCGLIFCONF) && defined(SIOCGLIFFLAGS) */
@@ -1201,7 +1198,7 @@ getlifaddrs2(struct ifaddrs **ifap,
 
 
 int ROKEN_LIB_FUNCTION
-getifaddrs(struct ifaddrs **ifap) 
+getifaddrs(struct ifaddrs **ifap)
 {
     int ret = -1;
     errno = ENXIO;
@@ -1241,14 +1238,14 @@ void ROKEN_LIB_FUNCTION
 freeifaddrs(struct ifaddrs *ifp)
 {
     struct ifaddrs *p, *q;
-    
+
     for(p = ifp; p; ) {
 	free(p->ifa_name);
 	if(p->ifa_addr)
 	    free(p->ifa_addr);
-	if(p->ifa_dstaddr) 
+	if(p->ifa_dstaddr)
 	    free(p->ifa_dstaddr);
-	if(p->ifa_netmask) 
+	if(p->ifa_netmask)
 	    free(p->ifa_netmask);
 	if(p->ifa_data)
 	    free(p->ifa_data);
@@ -1269,25 +1266,25 @@ print_addr(const char *s, struct sockaddr *sa)
     for(i = 0; i < sa->sa_len - ((long)sa->sa_data - (long)&sa->sa_family); i++)
 	printf("%02x", ((unsigned char*)sa->sa_data)[i]);
 #else
-    for(i = 0; i < sizeof(sa->sa_data); i++) 
+    for(i = 0; i < sizeof(sa->sa_data); i++)
 	printf("%02x", ((unsigned char*)sa->sa_data)[i]);
 #endif
     printf("\n");
 }
 
-void 
+void
 print_ifaddrs(struct ifaddrs *x)
 {
     struct ifaddrs *p;
-    
+
     for(p = x; p; p = p->ifa_next) {
 	printf("%s\n", p->ifa_name);
 	printf("  flags=%x\n", p->ifa_flags);
 	if(p->ifa_addr)
 	    print_addr("addr", p->ifa_addr);
-	if(p->ifa_dstaddr) 
+	if(p->ifa_dstaddr)
 	    print_addr("dstaddr", p->ifa_dstaddr);
-	if(p->ifa_netmask) 
+	if(p->ifa_netmask)
 	    print_addr("netmask", p->ifa_netmask);
 	printf("  %p\n", p->ifa_data);
     }
