@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2001, 2002, 2004, 2005, 2006, 2008, 2009, 2010
+ * Copyright (c) 2001, 2002, 2004, 2005, 2006, 2008, 2009, 2010, 2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -51,7 +51,7 @@
 #if HAVE_PAM
 
 static const char rcsid[] =
-"$Id: auth_pam.c,v 1.61.4.5 2010/09/21 11:24:43 karls Exp $";
+"$Id: auth_pam.c,v 1.78 2011/05/18 13:48:46 karls Exp $";
 
 static int
 pam_conversation(int msgc, const struct pam_message **msgv,
@@ -61,13 +61,6 @@ pam_conversation(int msgc, const struct pam_message **msgv,
  */
 
 
-typedef struct
-{
-   const char *user;
-   const char *password;
-} _pam_data_t;
-
-
 int
 pam_passwordcheck(s, src, dst, auth, emsg, emsgsize)
    int s;
@@ -100,8 +93,8 @@ pam_passwordcheck(s, src, dst, auth, emsg, emsgsize)
       { PAM_RUSER, "PAM_RUSER", DEFAULT_PAM_RUSER },
    };
 
-   slog(LOG_DEBUG, "%s: pam service name to use for user \"%s\": %s",
-   function, auth->name, auth->servicename);
+   slog(LOG_DEBUG, "%s: user \"%s\", servicename \"%s\", emsgsize %ld",
+   function, auth->name, auth->servicename, (long)emsgsize);
 
    /*
     * Note: we can not save the state of pam after pam_start(3), as
@@ -157,6 +150,9 @@ pam_passwordcheck(s, src, dst, auth, emsg, emsgsize)
    if ((rc = pam_authenticate(pamh, 0)) != PAM_SUCCESS) {
       sockd_priv(SOCKD_PRIV_PAM, PRIV_OFF);
 
+      slog(LOG_DEBUG, "%s: pam_authenticate() failed: %s",
+      function, pam_strerror(pamh, rc));
+
       snprintf(emsg, emsgsize, "pam_authenticate(): %s",
       pam_strerror(pamh, rc));
 
@@ -164,9 +160,14 @@ pam_passwordcheck(s, src, dst, auth, emsg, emsgsize)
       return -1;
    }
 
+   /* LINTED passing const, expecting non-const (PAM_SILENT) */
    rc = pam_acct_mgmt(pamh, PAM_SILENT);
    sockd_priv(SOCKD_PRIV_PAM, PRIV_OFF);
+
    if (rc != PAM_SUCCESS) {
+      slog(LOG_DEBUG, "%s: pam_acct_mgmt() failed: %s",
+      function, pam_strerror(pamh, rc));
+
       snprintf(emsg, emsgsize, "pam_acct_mgmt(): %s", pam_strerror(pamh, rc));
 
       pam_end(pamh, rc);
@@ -174,8 +175,10 @@ pam_passwordcheck(s, src, dst, auth, emsg, emsgsize)
    }
 
    if ((rc = pam_end(pamh, rc)) != PAM_SUCCESS)
-      swarnx("%s: strange ... pam_end() failed: %s", pam_strerror(pamh, rc));
+      swarnx("%s: strange ... pam_end() failed: %s",
+      function, pam_strerror(pamh, rc));
 
+   slog(LOG_DEBUG, "%s: pam authentication succeeded", function);
    return 0;
 }
 
@@ -196,7 +199,9 @@ pam_conversation(msgc, msgv, rspv, authdata)
    }
 
    if (((*rspv) = malloc(msgc * sizeof(struct pam_response))) == NULL) {
-      swarn("%s: malloc(%d * %d)", function, msgc, sizeof(struct pam_response));
+      swarn("%s: malloc(%d * %lu)",
+      function, msgc, (unsigned long)sizeof(struct pam_response));
+
       return PAM_CONV_ERR;
    }
 
@@ -222,7 +227,8 @@ pam_conversation(msgc, msgv, rspv, authdata)
             break;
 
          case PAM_ERROR_MSG:
-            slog(LOG_INFO, "%s: %s", function, msgv[i]->msg);
+            slog(LOG_INFO, "%s: got a pam error msg: %s",
+            function, msgv[i]->msg);
             break;
 
          case PAM_TEXT_INFO:
@@ -230,7 +236,8 @@ pam_conversation(msgc, msgv, rspv, authdata)
              * not expecting this, and where it has been seen (some versions
              * of FreeBSD), the string has been empty.
              */
-            slog(LOG_DEBUG, "%s: %s", function, msgv[i]->msg);
+            slog(LOG_DEBUG, "%s: got PAM_TEXT_INFO: \"%s\"",
+            function, msgv[i]->msg);
             break;
 
          default:
