@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009, 2010
+ * Copyright (c) 2009, 2010, 2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -51,62 +51,62 @@
 #if HAVE_GSSAPI
 
 static const char rcsid[] =
-   "$Id: method_gssapi.c,v 1.41.4.2 2010/09/21 11:24:43 karls Exp $";
+   "$Id: method_gssapi.c,v 1.55 2011/06/27 13:18:05 michaels Exp $";
 
-static int
+static negotiate_result_t
 recv_gssapi_auth_ver(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_auth_type(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_auth_len(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_auth_token(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_enc_ver(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_enc_type(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_enc_len(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_enc_token(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_packet(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_packet_ver(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_packet_type(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_packet_len(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_gssapi_packet_token(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
 
-int
+negotiate_result_t
 method_gssapi(s, request, state)
    int s;
    struct request_t *request;
@@ -115,7 +115,7 @@ method_gssapi(s, request, state)
 {
    const char *function = "method_gssapi()";
 
-   slog(LOG_DEBUG, function);
+   slog(LOG_DEBUG, "%s", function);
 
    request->auth->mdata.gssapi.state.id = GSS_C_NO_CONTEXT;
 
@@ -134,7 +134,7 @@ method_gssapi(s, request, state)
  *
  */
 
-static int
+static negotiate_result_t
 recv_gssapi_auth_ver(s, request, state)
    int s;
    struct request_t *request;
@@ -155,7 +155,7 @@ recv_gssapi_auth_ver(s, request, state)
          "%s: unknown version on gssapi packet from client: %d",
          function, gssapi_auth_version);
 
-         return -1;
+         return NEGOTIATE_ERROR;
    }
 
    state->rcurrent = recv_gssapi_auth_type;
@@ -163,7 +163,7 @@ recv_gssapi_auth_ver(s, request, state)
 }
 
 
-static int
+static negotiate_result_t
 recv_gssapi_auth_type(s, request, state)
    int s;
    struct request_t *request;
@@ -186,21 +186,21 @@ recv_gssapi_auth_type(s, request, state)
          "Got type %d, expected type %d",
          function, gssapi_auth_type, SOCKS_GSSAPI_AUTHENTICATION);
 
-         return -1;
+         return NEGOTIATE_ERROR;
 
       default:
          snprintf(state->emsg, sizeof(state->emsg),
          "%s: unknown type on gssapi packet from client: %d",
          function, gssapi_auth_type);
 
-         return -1;
+         return NEGOTIATE_ERROR;
    }
 
    state->rcurrent = recv_gssapi_auth_len;
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_gssapi_auth_len(s, request, state)
    int s;
    struct request_t *request;
@@ -216,7 +216,7 @@ recv_gssapi_auth_len(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_gssapi_auth_token(s, request, state)
    int s;
    struct request_t *request;
@@ -251,64 +251,51 @@ recv_gssapi_auth_token(s, request, state)
    gsskrb5_register_acceptor_identity(request->auth->mdata.gssapi.keytab);
 #endif /* HAVE_HEIMDAL_KERBEROS */
 
-   if (strcasecmp(request->auth->mdata.gssapi.servicename, "GSS_C_NO_NAME")
-   == 0) {
-      sockd_priv(SOCKD_PRIV_GSSAPI, PRIV_ON);
-      major_status = gss_acquire_cred(&minor_status, GSS_C_NO_NAME,
-                                      GSS_C_INDEFINITE,
-                                      GSS_C_NO_OID_SET, GSS_C_ACCEPT,
-                                      &server_creds,
-                                      NULL, NULL);
-      sockd_priv(SOCKD_PRIV_GSSAPI, PRIV_OFF);
+   slog(LOG_DEBUG,"%s: using gssapi service name %s",
+   function, request->auth->mdata.gssapi.servicename);
 
-      slog(LOG_DEBUG,"%s: using gssapi service name %s",
-      function, request->auth->mdata.gssapi.servicename);
-   }
-   else {
+   if (strcasecmp(request->auth->mdata.gssapi.servicename, "GSS_C_NO_NAME")
+   != 0) {
       gss_buffer_desc service;
 
-      service.length = strlen(request->auth->mdata.gssapi.servicename);
-      service.value  =  request->auth->mdata.gssapi.servicename;
+      service.value  = request->auth->mdata.gssapi.servicename;
+      service.length = strlen(service.value);
 
-      if (strchr(request->auth->mdata.gssapi.servicename, '/')) {
-         major_status = gss_import_name(&minor_status,
-                                        &service,
-                                        (gss_OID)GSS_C_NULL_OID,
-                                        &server_name);
-      }
-      else {
-         major_status = gss_import_name(&minor_status,
-                                        &service,
-                                        (gss_OID)gss_nt_service_name,
-                                        &server_name);
-      }
+      major_status
+      = gss_import_name(&minor_status,
+                        &service,
+                        strchr(request->auth->mdata.gssapi.servicename, '/')
+                        != NULL ?
+                                   (gss_OID)GSS_C_NULL_OID
+                                 : (gss_OID)gss_nt_service_name,
+                        &server_name);
 
       if (gss_err_isset(major_status, minor_status, emsg, sizeof(emsg))) {
          snprintf(state->emsg, sizeof(state->emsg),
          "%s: gss_import_name() %s", function, emsg);
 
          CLEAN_GSS_AUTH(client_name, server_name, server_creds);
-         return -1;
+         return NEGOTIATE_ERROR;
       }
-
-      slog(LOG_DEBUG,"%s: using gssapi service name %s",
-      function, request->auth->mdata.gssapi.servicename);
-
-      sockd_priv(SOCKD_PRIV_GSSAPI, PRIV_ON);
-      major_status = gss_acquire_cred(&minor_status, server_name,
-                                      GSS_C_INDEFINITE,
-                                      GSS_C_NO_OID_SET, GSS_C_ACCEPT,
-                                      &server_creds,
-                                      NULL, NULL);
-      sockd_priv(SOCKD_PRIV_GSSAPI, PRIV_OFF);
    }
 
-   if (gss_err_isset(major_status, minor_status, emsg, sizeof(emsg))) {
-      snprintf(state->emsg, sizeof(state->emsg),
-      "%s: gss_acquire_cred(): %s", function, emsg);
+   sockd_priv(SOCKD_PRIV_GSSAPI, PRIV_ON);
+   major_status = gss_acquire_cred(&minor_status,
+                                   server_name,
+                                   GSS_C_INDEFINITE,
+                                   GSS_C_NO_OID_SET,
+                                   GSS_C_ACCEPT,
+                                   &server_creds,
+                                   NULL,
+                                   NULL);
+   sockd_priv(SOCKD_PRIV_GSSAPI, PRIV_OFF);
 
-       CLEAN_GSS_AUTH(client_name, server_name, server_creds);
-       return -1;
+   if (gss_err_isset(major_status, minor_status, emsg, sizeof(emsg))) {
+      snprintf(state->emsg, sizeof(state->emsg), "%s: gss_acquire_cred(): %s",
+               function, emsg);
+
+      CLEAN_GSS_AUTH(client_name, server_name, server_creds);
+      return NEGOTIATE_ERROR;
    }
 
    sockd_priv(SOCKD_PRIV_GSSAPI, PRIV_ON);
@@ -326,12 +313,15 @@ recv_gssapi_auth_token(s, request, state)
                             NULL);
    sockd_priv(SOCKD_PRIV_GSSAPI, PRIV_OFF);
 
+   slog(LOG_DEBUG, "%s: length of output_token is %lu",
+   function, (unsigned long)output_token.length);
+
    if (gss_err_isset(major_status, minor_status, emsg, sizeof(emsg))) {
       snprintf(state->emsg, sizeof(state->emsg),
       "%s: gss_accept_sec_context() failed: %s", function, emsg);
 
-       CLEAN_GSS_AUTH(client_name, server_name, server_creds);
-       return -1;
+      CLEAN_GSS_AUTH(client_name, server_name, server_creds);
+      return NEGOTIATE_ERROR;
    }
 
    /*
@@ -368,11 +358,14 @@ recv_gssapi_auth_token(s, request, state)
    if ((rc = socks_sendton(s, buf, buflen, 0, 0, NULL, 0, request->auth))
    != (ssize_t)buflen)  {
       snprintf(state->emsg, sizeof(state->emsg),
-      "socks_sendton() token: wrote %ld out of %lu byte%s: %s",
-      (long)rc, (unsigned long)buflen, buflen == 1 ? "" : "s", strerror(errno));
+              "socks_sendton() token: wrote %ld out of %lu byte%s: %s",
+               (long)rc,
+               (unsigned long)buflen,
+               buflen == 1 ? "" : "s",
+               strerror(errno));
 
       CLEAN_GSS_AUTH(client_name, server_name, server_creds);
-      return -1;
+      return NEGOTIATE_ERROR;
    }
 
    if (major_status == GSS_S_COMPLETE) {
@@ -385,7 +378,7 @@ recv_gssapi_auth_token(s, request, state)
          "%s: gss_display_name(): %s", function, emsg);
 
          CLEAN_GSS_AUTH(client_name, server_name, server_creds);
-         return -1;
+         return NEGOTIATE_ERROR;
       }
 
       memcpy(request->auth->mdata.gssapi.name, output_token.value,
@@ -401,6 +394,7 @@ recv_gssapi_auth_token(s, request, state)
    else if (major_status == GSS_S_CONTINUE_NEEDED) {
       CLEAN_GSS_AUTH(client_name, server_name, server_creds);
 
+      /* expect a new token, with the version, length, etc. header. */
       state->rcurrent = recv_gssapi_auth_ver;
       return state->rcurrent(s, request, state);
    }
@@ -409,13 +403,13 @@ recv_gssapi_auth_token(s, request, state)
       "%s: unknown gss major_status %d", function, major_status);
 
       CLEAN_GSS_AUTH(client_name, server_name, server_creds);
-      return -1;
+      return NEGOTIATE_ERROR;
    }
 
    /* NOTREACHED */
 }
 
-static int
+static negotiate_result_t
 recv_gssapi_enc_ver(s, request, state)
    int s;
    struct request_t *request;
@@ -436,7 +430,7 @@ recv_gssapi_enc_ver(s, request, state)
          "%s: unknown version on gssapi packet from client: %d",
          function, gssapi_enc_version);
 
-         return -1;
+         return NEGOTIATE_ERROR;
    }
 
    state->rcurrent = recv_gssapi_enc_type;
@@ -453,7 +447,7 @@ recv_gssapi_enc_ver(s, request, state)
  *   +------+------+------+.......................+
  *
  */
-static int
+static negotiate_result_t
 recv_gssapi_enc_type(s, request, state)
    int s;
    struct request_t *request;
@@ -475,21 +469,21 @@ recv_gssapi_enc_type(s, request, state)
           "%s: unsupported per message encryption on gssapi packet from client",
           function);
 
-         return -1;
+         return NEGOTIATE_ERROR;
 
       default:
          snprintf(state->emsg, sizeof(state->emsg),
          "%s: unknown type on gssapi packet from client: %d",
          function, gssapi_enc_type);
 
-         return -1;
+         return NEGOTIATE_ERROR;
    }
 
    state->rcurrent = recv_gssapi_enc_len;
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_gssapi_enc_len(s, request, state)
    int s;
    struct request_t *request;
@@ -506,7 +500,7 @@ recv_gssapi_enc_len(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_gssapi_enc_token(s, request, state)
    int s;
    struct request_t *request;
@@ -548,27 +542,24 @@ recv_gssapi_enc_token(s, request, state)
                    request->auth->mdata.gssapi.state.id,
                    &input_token, &output_token, 0, GSS_C_QOP_DEFAULT);
 
-      if (output_token.length != 1) {
-         snprintf(state->emsg, sizeof(state->emsg),
-         "%s: gssapi encryption unwrapped length is wrong: %lu, expected 1.  "
-         "Is the client a NEC-based client perhaps?  If so, try enabling NEC "
-         "client-compatibility in this client-rule ",
-         function, (unsigned long)output_token.length);
-
-         CLEAN_GSS_TOKEN(output_token);
-
-         errno = EPROTO;
-         return -1;
-      }
-
       if (gss_err_isset(major_status, minor_status, emsg, sizeof(emsg))) {
          snprintf(state->emsg, sizeof(state->emsg),
          "%s: gss_unwrap(): %s", function, emsg);
 
          CLEAN_GSS_TOKEN(output_token);
+         return NEGOTIATE_ERROR;
+      }
 
-         errno = EPROTO;
-         return -1;
+      if (output_token.length != 1) {
+         snprintf(state->emsg, sizeof(state->emsg),
+                  "%s: gssapi encryption unwrapped length is wrong: is %lu, "
+                  "but expected is 1.  Is the client a NEC-based client "
+                  "perhaps?  If so, try enabling NEC client-compatibility for "
+                  "this client-rule",
+                  function, (unsigned long)output_token.length);
+
+         CLEAN_GSS_TOKEN(output_token);
+         return NEGOTIATE_ERROR;
       }
 
       SASSERTX(output_token.length == sizeof(c));
@@ -606,8 +597,7 @@ recv_gssapi_enc_token(s, request, state)
       request->auth->mdata.gssapi.encryption.confidentiality,
       request->auth->mdata.gssapi.encryption.permessage);
 
-      errno = EPROTO;
-      return -1;
+      return NEGOTIATE_ERROR;
    }
 
    gss_server_enc = gss_enc;
@@ -632,9 +622,7 @@ recv_gssapi_enc_token(s, request, state)
          "%s: gss_wrap(): %s", function, emsg);
 
          CLEAN_GSS_TOKEN(output_token);
-
-         errno = EPROTO;
-         return -1;
+         return NEGOTIATE_ERROR;
       }
 
       token_length = htons(output_token.length);
@@ -672,12 +660,14 @@ recv_gssapi_enc_token(s, request, state)
    if ((rc = socks_sendton(s, buf, buflen, 0, 0, NULL, 0, request->auth))
    != (ssize_t)buflen) {
       snprintf(state->emsg, sizeof(state->emsg),
-      "socks_sendton() buf: wrote %ld out of %lu byte%s: %s",
-      (long)rc, (unsigned long)buflen, buflen == 1 ? "" : "s",
-      strerror(errno));
+               "socks_sendton() buf: wrote %ld out of %lu byte%s: %s",
+                (long)rc,
+                (unsigned long)buflen,
+                buflen == 1 ? "" : "s",
+                strerror(errno));
 
       CLEAN_GSS_TOKEN(output_token);
-      return -1;
+      return NEGOTIATE_ERROR;
    }
 
    CLEAN_GSS_TOKEN(output_token);
@@ -704,7 +694,7 @@ recv_gssapi_enc_token(s, request, state)
  *   +------+------+------+.......................+
  *
  */
-static int
+static negotiate_result_t
 recv_gssapi_packet(s, request, state)
    int s;
    struct request_t *request;
@@ -716,7 +706,7 @@ recv_gssapi_packet(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_gssapi_packet_ver(s, request, state)
    int s;
    struct request_t *request;
@@ -736,15 +726,15 @@ recv_gssapi_packet_ver(s, request, state)
          snprintf(state->emsg, sizeof(state->emsg),
          "%s: unknown version on gssapi packet from client: %d",
          function, gssapi_packet_version);
-         errno = EPROTO;
-         return -1;
+
+         return NEGOTIATE_ERROR;
    }
 
    state->rcurrent = recv_gssapi_packet_type;
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_gssapi_packet_type(s, request, state)
    int s;
    struct request_t *request;
@@ -766,23 +756,21 @@ recv_gssapi_packet_type(s, request, state)
          "%s: received out of sequence exchange from client, type: %d",
          function, gssapi_enc_type);
 
-         errno = EPROTO;
-         return -1;
+         return NEGOTIATE_ERROR;
 
       default:
          snprintf(state->emsg, sizeof(state->emsg),
          "%s: unknown type on gssapi packet from client: %d",
          function, gssapi_enc_type);
 
-         errno = EPROTO;
-         return -1;
+         return NEGOTIATE_ERROR;
    }
 
    state->rcurrent = recv_gssapi_packet_len;
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_gssapi_packet_len(s, request, state)
    int s;
    struct request_t *request;
@@ -799,7 +787,7 @@ recv_gssapi_packet_len(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_gssapi_packet_token(s, request, state)
    int s;
    struct request_t *request;
@@ -833,8 +821,7 @@ recv_gssapi_packet_token(s, request, state)
       snprintf(state->emsg, sizeof(state->emsg),
       "%s: gss_unwrap(): %s", function, emsg);
 
-      errno = EPROTO;
-      return -1;
+      return NEGOTIATE_ERROR;
    }
 
    offset = 0;
@@ -846,9 +833,7 @@ recv_gssapi_packet_token(s, request, state)
       function, (unsigned long)output_token.length);
 
       CLEAN_GSS_TOKEN(output_token);
-
-      errno = EPROTO;
-      return -1;
+      return NEGOTIATE_ERROR;
    }
    memcpy(&request->version, &data[offset], sizeof(request->version));
    if (request->version != PROXY_SOCKS_V5) {
@@ -856,9 +841,7 @@ recv_gssapi_packet_token(s, request, state)
       "%s: invalid socks version %d in request", function, request->version);
 
       CLEAN_GSS_TOKEN(output_token);
-
-      errno = EPROTO;
-      return -1;
+      return NEGOTIATE_ERROR;
    }
    offset += sizeof(request->version);
 
@@ -868,9 +851,7 @@ recv_gssapi_packet_token(s, request, state)
       function, (unsigned long)output_token.length);
 
       CLEAN_GSS_TOKEN(output_token);
-
-      errno = EPROTO;
-      return -1;
+      return NEGOTIATE_ERROR;
    }
    memcpy(&request->command, &data[offset], sizeof(request->command));
    offset += sizeof(request->command);
@@ -891,14 +872,12 @@ recv_gssapi_packet_token(s, request, state)
          function, request->command);
 
          CLEAN_GSS_TOKEN(output_token);
-
-         errno = EPROTO;
-         return -1;
+         return NEGOTIATE_ERROR;
    }
 
    if (offset + sizeof(request->flag) > output_token.length) {
       CLEAN_GSS_TOKEN(output_token);
-      return -1;
+      return NEGOTIATE_ERROR;
    }
    memcpy(&request->flag, &data[offset], sizeof(request->flag));
    offset += sizeof(request->flag);
@@ -906,16 +885,16 @@ recv_gssapi_packet_token(s, request, state)
    if (mem2sockshost(&request->host, &data[offset],
    output_token.length - offset, request->version) == NULL) {
       CLEAN_GSS_TOKEN(output_token);
-      return -1;
+      return NEGOTIATE_ERROR;
    }
 
    CLEAN_GSS_TOKEN(output_token);
 
    /* Negotiation finished => set connection state as protected */
    if (request->auth->mdata.gssapi.state.protection)
-      request->auth->mdata.gssapi.state.encryption = GSSAPI_ENCRYPT;
+      request->auth->mdata.gssapi.state.wrap = 1;
 
-   return 1;
+   return NEGOTIATE_FINISHED;
 }
 
 #endif /* HAVE_GSSAPI */
