@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2005, 2008, 2009, 2010
+ * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2005, 2008, 2009, 2010,
+ *               2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -44,67 +45,79 @@
 #include "common.h"
 
 static const char rcsid[] =
-"$Id: sockd_protocol.c,v 1.134.4.2 2010/09/21 11:24:43 karls Exp $";
+"$Id: sockd_protocol.c,v 1.167 2011/07/08 10:11:52 michaels Exp $";
 
-static int
+#if SOCKS_SERVER
+static negotiate_result_t
 recv_v4req(int s, struct request_t *request, struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_v5req(int s, struct request_t *request, struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_methods(int s, struct request_t *request,
              struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_ver(int s, struct request_t *request, struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_cmd(int s, struct request_t *request, struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_flag(int s, struct request_t *request, struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_sockshost(int s, struct request_t *request,
       struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_atyp(int s, struct request_t *request, struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_port(int s, struct request_t *request, struct negotiate_state_t *state);
 
-static int
-recv_address(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+static negotiate_result_t
+recv_address(int s, struct request_t *request, struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_domain(int s, struct request_t *request, struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 recv_username(int s, struct request_t *request,
-      struct negotiate_state_t *state);
+              struct negotiate_state_t *state);
 
-static int
+static negotiate_result_t
 methodnegotiate(int s, struct request_t *request,
       struct negotiate_state_t *state);
+#endif /* SOCKS_SERVER */
 
-
-int
-recv_request(s, request, state)
+negotiate_result_t
+recv_clientrequest(s, request, state)
    int s;
    struct request_t *request;
    struct negotiate_state_t *state;
 {
-   int rc;
+   const char *function = "recv_clientrequest()";
+#if HAVE_NEGOTIATE_PHASE
+   negotiate_result_t rc;
+#endif /* HAVE_NEGOTIATE_PHASE */
+
+   slog(LOG_DEBUG, "%s: client %s, read so far: %lu",
+                   function,
+                   sockshost2string(&state->src, NULL, 0),
+                   (unsigned long)state->reqread);
 
    if (state->complete)
-      return 1;
+      return NEGOTIATE_FINISHED;
 
+#if HAVE_NEGOTIATE_PHASE
    if (state->rcurrent != NULL)   /* not first call on this client. */
       rc = state->rcurrent(s, request, state);
    else {
+      char src[MAXSOCKSHOSTSTRING], dst[sizeof(src)];
+
+#if SOCKS_SERVER
       INIT(sizeof(request->version));
       CHECK(&request->version, request->auth, NULL);
 
@@ -118,19 +131,174 @@ recv_request(s, request, state)
             break;
 
          default:
-            slog(LOG_DEBUG, "unknown version %d in request", request->version);
-            return -1;
+            snprintf(state->emsg, sizeof(state->emsg),
+                     "unknown socks version %d in clientrequest",
+                     request->version);
+            return NEGOTIATE_ERROR;
       }
 
+#elif COVENANT
+      state->rcurrent = recv_httprequest;
+#else /* !COVENANT */
+      SASSERTX(0); /* should never have been called. */
+#endif
+
+      slog(LOG_DEBUG, "%s: initiating negotiation with client at %s "
+                      "which connected to us on %s",
+                      function,
+                      sockshost2string(&state->src, src, sizeof(src)),
+                      sockshost2string(&state->dst, dst, sizeof(dst)));
+
+
       rc = state->rcurrent(s, request, state);
    }
 
-   state->complete = rc > 0; /* complete request read? */
-
+   state->complete = (rc == NEGOTIATE_FINISHED);
    return rc;
+#else /* !HAVE_NEGOTIATE_PHASE */
+
+   SASSERTX(state->complete);
+   /* NOTREACHED */
+
+#endif /* !HAVE_NEGOTIATE_PHASE */
+}
+
+#if HAVE_NEGOTIATE_PHASE
+void
+send_failure(s, response, failure)
+   int s;
+   struct response_t *response;
+   int failure;
+{
+#if HAVE_GSSAPI
+   const char *function = "send_failure()";
+   gss_buffer_desc output_token;
+   OM_uint32 minor_status;
+#endif /* HAVE_GSSAPI */
+
+   socks_set_responsevalue(response, sockscode(response->version, failure));
+   send_response(s, response);
+
+#if HAVE_GSSAPI
+   if (response->auth->method == AUTHMETHOD_GSSAPI) {
+      if (gss_delete_sec_context(&minor_status,
+         &response->auth->mdata.gssapi.state.id, &output_token)
+         != GSS_S_COMPLETE)
+            swarn("%s: gss_delete_sec_context failed", function);
+
+      CLEAN_GSS_TOKEN(output_token);
+   }
+#endif /* HAVE_GSSAPI */
 }
 
 int
+send_response(s, response)
+   int s;
+   const struct response_t *response;
+{
+   const char *function = "send_response()";
+   ssize_t sendt;
+   size_t length;
+#if SOCKS_SERVER
+   unsigned char responsemem[sizeof(*response)], *p = responsemem;
+#else
+   char responsemem[1024], *p = responsemem;
+#endif
+
+   switch (response->version) {
+#if SOCKS_SERVER
+      case PROXY_SOCKS_V4REPLY_VERSION:
+         /*
+          * socks V4 reply packet:
+          *
+          *  VN   CD  DSTPORT  DSTIP
+          *  1  + 1  +  2    +  4
+          *
+          *  Always 8 octets long.
+          */
+
+         memcpy(p, &response->version, sizeof(response->version));
+         p += sizeof(response->version);
+
+         /* CD (reply) */
+         memcpy(p, &response->reply.socks, sizeof(response->reply.socks));
+         p += sizeof(response->reply.socks);
+
+         p      = sockshost2mem(&response->host, p, response->version);
+         length = p - responsemem;
+
+         break;
+
+      case PROXY_SOCKS_V5:
+         /*
+          * socks V5 reply:
+          *
+          * +----+-----+-------+------+----------+----------+
+          * |VER | REP |  FLAG | ATYP | BND.ADDR | BND.PORT |
+          * +----+-----+-------+------+----------+----------+
+          * | 1  |  1  |   1   |  1   | Variable |    2     |
+          * +----+-----+-------+------+----------+----------+
+          *   1     1      1      1                   2
+          *
+          * Which gives a fixed size of at least 6 octets.
+          * The first octet of DST.ADDR when it is SOCKS_ADDR_DOMAINNAME
+          * contains the length.
+          *
+          */
+
+         /* VER */
+         memcpy(p, &response->version, sizeof(response->version));
+         p += sizeof(response->version);
+
+         /* REP */
+         memcpy(p, &response->reply.socks, sizeof(response->reply.socks));
+         p += sizeof(response->reply.socks);
+
+         /* FLAG */
+         memcpy(p, &response->flag, sizeof(response->flag));
+         p += sizeof(response->flag);
+
+         p = sockshost2mem(&response->host, p, response->version);
+         break;
+
+#elif COVENANT /* !SOCKS_SERVER */
+      case PROXY_HTTP_10:
+      case PROXY_HTTP_11: {
+         size_t l;
+
+         l  = httpresponse2mem(s, response, responsemem, sizeof(responsemem));
+         p += l;
+
+         break;
+      }
+#endif /* COVENANT */
+
+      default:
+         SERRX(response->version);
+   }
+
+   length = p - responsemem;
+
+#if SOCKS_SERVER
+   slog(LOG_DEBUG, "%s: sending response: %s",
+   function, socks_packet2string(response, 0));
+#else /* COVENANT */
+   slog(LOG_DEBUG, "%s: sending response:\n%s", function, responsemem);
+#endif
+
+   if ((sendt = socks_sendton(s, responsemem, length, 1, 0, NULL, 0,
+   response->auth)) != (ssize_t)length) {
+      slog(LOG_DEBUG, "%s: socks_sendton(): %ld/%lu: %s",
+      function, (long)sendt, (long unsigned)length, strerror(errno));
+
+      return -1;
+   }
+
+   return 0;
+}
+
+#if SOCKS_SERVER
+negotiate_result_t
 recv_sockspacket(s, request, state)
    int s;
    struct request_t *request;
@@ -140,7 +308,7 @@ recv_sockspacket(s, request, state)
    return recv_ver(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_v4req (s, request, state)
    int s;
    struct request_t *request;
@@ -161,7 +329,7 @@ recv_v4req (s, request, state)
 
 }
 
-static int
+static negotiate_result_t
 recv_v5req (s, request, state)
    int s;
    struct request_t *request;
@@ -203,11 +371,10 @@ recv_v5req (s, request, state)
    OCTETIFY(state->mem[start]);
 
    state->rcurrent = recv_methods;
-
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_methods(s, request, state)
    int s;
    struct request_t *request;
@@ -227,25 +394,37 @@ recv_methods(s, request, state)
    CHECK(&state->mem[start], request->auth, NULL);
 
    for (i = bufused = 0; i < methodc; ++i)
-      bufused += snprintfn(&buf[bufused], sizeof(buf) - bufused, "0x%x, ",
+      bufused += snprintf(&buf[bufused], sizeof(buf) - bufused, "0x%x, ",
       state->mem[start + i]);
 
-   STRIPTRAILING(buf, bufused);
+   STRIPTRAILING(buf, bufused, ", \t\n");
 
    slog(LOG_DEBUG, "%s: client offered %d authentication method%s: %s",
    function, methodc, methodc == 1 ? "" : "s", buf);
 
    switch (request->auth->method) {
-      case AUTHMETHOD_GSSAPI: {
+      case AUTHMETHOD_NOTSET:
+         slog(LOG_DEBUG, "%s: method set to %d, selecting new",
+         function, request->auth->method);
+
+         request->auth->method = selectmethod(sockscf.methodv, sockscf.methodc,
+         &state->mem[start], (size_t)methodc);
+
+         break;
+
+      default: {
          /*
           * Socks-methods that can be decided for use before we receive
-          * the actual request.  So far only gssapi.
+          * the actual request.  Normally only gssapi, but if the
+          * rule has singleauth enabled and the client matches the
+          * criteria for it, the socks-method will also have been
+          * chosen already (should be NONE).
           */
          size_t i;
 
          slog(LOG_DEBUG, "%s: method %d already chosen for this rule, "
                          "not selecting again",
-         function, request->auth->method);
+                         function, request->auth->method);
 
          for (i = 0; i < methodc; ++i)
             if (state->mem[start + i] == request->auth->method)
@@ -256,15 +435,6 @@ recv_methods(s, request, state)
 
          break;
       }
-
-      default:
-         slog(LOG_DEBUG, "%s: method set to %d, selecting new",
-         function, request->auth->method);
-
-         request->auth->method = selectmethod(sockscf.methodv, sockscf.methodc,
-         &state->mem[start], (size_t)methodc);
-
-         break;
    }
 
    /* send reply:
@@ -284,20 +454,20 @@ recv_methods(s, request, state)
 
    if (socks_sendton(s, reply, sizeof(reply), sizeof(reply), 0, NULL,
    0, request->auth) != sizeof(reply))
-      return -1;
+      return NEGOTIATE_ERROR;
 
    if (request->auth->method == AUTHMETHOD_NOACCEPT) {
       snprintf(state->emsg, sizeof(state->emsg),
       "client offered no acceptable authentication method");
-      errno = EPROTO;
-      return -1;
+
+      return NEGOTIATE_ERROR;
    }
 
    state->rcurrent = methodnegotiate;
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 methodnegotiate(s, request, state)
    int s;
    struct request_t *request;
@@ -327,7 +497,7 @@ methodnegotiate(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_ver(s, request, state)
    int s;
    struct request_t *request;
@@ -346,7 +516,7 @@ recv_ver(s, request, state)
 
          default:
             slog(LOG_DEBUG, "unknown version %d in request", request->version);
-            return -1;
+            return NEGOTIATE_ERROR;
       }
    }
 
@@ -354,7 +524,7 @@ recv_ver(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_cmd(s, request, state)
    int s;
    struct request_t *request;
@@ -376,6 +546,7 @@ recv_cmd(s, request, state)
          break;
 
       default:
+         /*  XXX should print to emsg, not swarn() here. */
          swarnx("%s: unknown command: %d", function, request->command);
    }
 
@@ -395,7 +566,7 @@ recv_cmd(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_flag(s, request, state)
    int s;
    struct request_t *request;
@@ -408,7 +579,7 @@ recv_flag(s, request, state)
    SERRX(0); /* NOTREACHED */
 }
 
-static int
+static negotiate_result_t
 recv_sockshost(s, request, state)
    int s;
    struct request_t *request;
@@ -430,7 +601,7 @@ recv_sockshost(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_atyp(s, request, state)
    int s;
    struct request_t *request;
@@ -443,7 +614,7 @@ recv_atyp(s, request, state)
    SERRX(0); /* NOTREACHED */
 }
 
-static int
+static negotiate_result_t
 recv_address(s, request, state)
    int s;
    struct request_t *request;
@@ -454,7 +625,8 @@ recv_address(s, request, state)
       case PROXY_SOCKS_V4: {
          INIT(sizeof(request->host.addr.ipv4));
 
-         request->host.atype = SOCKS_ADDR_IPV4; /* only one supported in v4. */
+         /* only one supported in v4. */
+         request->host.atype = (unsigned char)SOCKS_ADDR_IPV4;
 
          CHECK(&request->host.addr.ipv4, request->auth, recv_username);
          SERRX(0); /* NOTREACHED */
@@ -488,7 +660,7 @@ recv_address(s, request, state)
             default:
                slog(LOG_DEBUG, "unknown address format %d in reply",
                request->host.atype);
-               return -1;
+               return NEGOTIATE_ERROR;
          }
 
       default:
@@ -498,7 +670,7 @@ recv_address(s, request, state)
    SERRX(0); /* NOTREACHED */
 }
 
-static int
+static negotiate_result_t
 recv_domain(s, request, state)
    int s;
    struct request_t *request;
@@ -520,7 +692,7 @@ recv_domain(s, request, state)
    return state->rcurrent(s, request, state);
 }
 
-static int
+static negotiate_result_t
 recv_port(s, request, state)
    int s;
    struct request_t *request;
@@ -536,7 +708,7 @@ recv_port(s, request, state)
          return state->rcurrent(s, request, state);
 
       case PROXY_SOCKS_V5:
-         return 1;   /* all done. */
+         return NEGOTIATE_FINISHED;   /* all done. */
 
       default:
          SERRX(request->version);
@@ -545,7 +717,7 @@ recv_port(s, request, state)
    SERRX(0); /* NOTREACHED */
 }
 
-static int
+static negotiate_result_t
 recv_username(s, request, state)
    int s;
    struct request_t *request;
@@ -572,11 +744,11 @@ recv_username(s, request, state)
 
          state->mem[state->reqread - 1] = NUL;
 
-         swarnx("%s: username too long (> %ld): \"%s\"", function,
+         swarnx("%s: username too long (> %ld): \"%s...\"", function,
          (unsigned long)strlen(username),
          str2vis(username, strlen(username), visstring, sizeof(visstring)));
 
-         return -1;
+         return NEGOTIATE_ERROR;
       }
 
       CHECK(&state->mem[start], request->auth, NULL);
@@ -585,8 +757,8 @@ recv_username(s, request, state)
        * Since we don't know how long the username is, we can only read one
        * byte at a time.  We don't want CHECK() to set state->rcurrent to
        * NULL after each successful read of that one byte, since
-       * recv_request() will then think we are starting from the beginning
-       * next time we call it.
+       * recv_clientrequest() will then think we are starting from the
+       * beginning next time we call it.
        */
       state->rcurrent = recv_username;
    } while (state->mem[state->reqread - 1] != 0);
@@ -595,88 +767,7 @@ recv_username(s, request, state)
    slog(LOG_DEBUG, "%s: got socks v4 username: \"%s\"", function, username);
 
    state->rcurrent = NULL;
-   return 1;   /* end of request. */
-}
-
-int
-send_response(s, response)
-   int s;
-   const struct response_t *response;
-{
-#if !BAREFOOTD
-   const char *function = "send_response()";
-   ssize_t sendt;
-   size_t length;
-   unsigned char responsemem[sizeof(*response)];
-   unsigned char *p = responsemem;
-
-   switch (response->version) {
-      case PROXY_SOCKS_V4REPLY_VERSION:
-         /*
-          * socks V4 reply packet:
-          *
-          *  VN   CD  DSTPORT  DSTIP
-          *  1  + 1  +  2    +  4
-          *
-          *  Always 8 octets long.
-          */
-
-         memcpy(p, &response->version, sizeof(response->version));
-         p += sizeof(response->version);
-
-         /* CD (reply) */
-         memcpy(p, &response->reply, sizeof(response->reply));
-         p += sizeof(response->reply);
-
-         break;
-
-      case PROXY_SOCKS_V5:
-         /*
-          * socks V5 reply:
-          *
-          * +----+-----+-------+------+----------+----------+
-          * |VER | REP |  FLAG | ATYP | BND.ADDR | BND.PORT |
-          * +----+-----+-------+------+----------+----------+
-          * | 1  |  1  |   1   |  1   | Variable |    2     |
-          * +----+-----+-------+------+----------+----------+
-          *   1     1      1      1                   2
-          *
-          * Which gives a fixed size of at least 6 octets.
-          * The first octet of DST.ADDR when it is SOCKS_ADDR_DOMAINNAME
-          * contains the length.
-          *
-          */
-
-         /* VER */
-         memcpy(p, &response->version, sizeof(response->version));
-         p += sizeof(response->version);
-
-         /* REP */
-         memcpy(p, &response->reply, sizeof(response->reply));
-         p += sizeof(response->reply);
-
-         /* FLAG */
-         memcpy(p, &response->flag, sizeof(response->flag));
-         p += sizeof(response->flag);
-
-         break;
-
-      default:
-         SERRX(response->version);
-   }
-
-   p = sockshost2mem(&response->host, p, response->version);
-   length = p - responsemem;
-
-   slog(LOG_DEBUG, "%s: sending response: %s",
-   function, socks_packet2string(response, SOCKS_RESPONSE));
-
-   if ((sendt = socks_sendton(s, responsemem, length, 1, 0, NULL, 0,
-   response->auth)) != (ssize_t)length) {
-      slog(LOG_DEBUG, "%s: socks_sendton(): %ld/%lu: %s",
-      function, (long)sendt, (long unsigned)length, strerror(errno));
-      return -1;
-   }
-#endif /* !BAREFOOTD */
-   return 0;
+   return NEGOTIATE_FINISHED;   /* end of request. */
 }
+#endif /* SOCKS_SERVER */
+#endif /* HAVE_NEGOTIATE_PHASE */
