@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2001, 2002, 2003
+ * Copyright (c) 2001, 2003, 2005, 2008, 2009, 2010, 2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -46,57 +46,65 @@
 #include <math.h> /* XXX */
 
 static const char rcsid[] =
-"$Id: bandwidth.c,v 1.12 2005/11/02 12:11:28 michaels Exp $";
+"$Id: bandwidth.c,v 1.27 2011/07/11 12:08:34 michaels Exp $";
 
 const char module_bandwidth_version[] =
-"$Id: bandwidth.c,v 1.12 2005/11/02 12:11:28 michaels Exp $";
+"$Id: bandwidth.c,v 1.27 2011/07/11 12:08:34 michaels Exp $";
 
 
-int
-bw_use(bw)
-	bw_t *bw;
+void
+bw_use(bw, lock)
+   shmem_object_t *bw;
+   const int lock;
 {
-	return 1;
-}
 
-bw_t *
-bw_alloc(client, number)
-	int client;
-	int number;
-{
-	return NULL;
+   (void)bw;
 }
 
 void
-bw_unuse(bw)
-	bw_t *bw;
+bw_unuse(bw, lock)
+   shmem_object_t *bw;
+   const int lock;
 {
 
+   (void)bw;
 }
 
 ssize_t
-bw_left(bw)
-	const bw_t *bw;
+bw_left(bw, lock)
+   const shmem_object_t *bw;
+   const int lock;
 {
 
-	return MAX(SOCKD_BUFSIZETCP, SOCKD_BUFSIZEUDP);
+   (void)bw;
+   return SOCKD_BUFSIZE;
 }
 
 void
-bw_update(bw, bwused, bwusedtime)
-	bw_t *bw;
-	size_t bwused;
-	const struct timeval *bwusedtime;
+bw_update(bw, bwused, bwusedtime, lock)
+   shmem_object_t *bw;
+   size_t bwused;
+   const struct timeval *bwusedtime;
+   const int lock;
 {
 
+   (void)bw;
+   (void)bwused;
+   (void)bwusedtime;
 }
 
-struct timeval *
-bw_isoverflow(bw, timenow, overflow)
-	bw_t *bw;
-	const struct timeval *timenow;
-	struct timeval *overflow;
+
+
+
+int
+bw_rulehasoverflown(rule, tnow, overflowok, controlfd, srcfd, dstfd)
+   const struct rule_t *rule;
+   const struct timeval *tnow;
+   struct timeval *overflowok;
+   const int controlfd;
+   const int srcfd;
+   const int dstfd;
 {
 
-	return NULL;
+   return 0;
 }
