@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005, 2006, 2008, 2009
+ * Copyright (c) 2005, 2006, 2008, 2009, 2010, 2011
  *      Inferno Nettverk A/S, Norway.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -39,30 +39,37 @@
  * any improvements or extensions that they make and grant Inferno Nettverk A/S
  * the rights to redistribute these changes.
  *
- * $Id: session.c,v 1.7 2009/10/22 17:36:13 karls Exp $
+ * $Id: session.c,v 1.12 2011/05/18 13:48:47 karls Exp $
  */
 
 #include "common.h"
 
-static const char rcsid[] = 
-"$Id: session.c,v 1.7 2009/10/22 17:36:13 karls Exp $";
+static const char rcsid[] =
+"$Id: session.c,v 1.12 2011/05/18 13:48:47 karls Exp $";
 
 const char module_session_version[] =
-"$Id: session.c,v 1.7 2009/10/22 17:36:13 karls Exp $";
+"$Id: session.c,v 1.12 2011/05/18 13:48:47 karls Exp $";
 
 extern const char *module_session_ipaddrv[];
 extern const int module_session_ipaddrc;
 
 void
-session_unuse(ss)
-   session_t *ss;
+session_unuse(ss, lock)
+   shmem_object_t *ss;
+   const int lock;
 {
+
+   (void)ss;
 }
 
 
 int
-session_use(ss)
-   session_t *ss;
+session_use(ss, lock)
+   shmem_object_t *ss;
+   const int lock;
 {
+
+   (void)ss;
+
    return 1;
 }
